/***************************************************************************
 *   Copyright (C) 2007, 2008, 2009 by Thomas Sailer                       *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <gtkmm/main.h>
#include <glibmm/optionentry.h>
#include <glibmm/optiongroup.h>
#include <glibmm/optioncontext.h>

#include "dbeditor.h"
#include "sysdepsgui.h"

AirspaceEditor::AirspaceModelColumns::AirspaceModelColumns( void )
{
        add(m_col_id);
        add(m_col_labellon);
        add(m_col_labellat);
        add(m_col_label_placement);
        add(m_col_sourceid);
        add(m_col_icao);
        add(m_col_name);
        add(m_col_ident);
        add(m_col_commname);
        add(m_col_commfreq0);
        add(m_col_commfreq1);
        add(m_col_classtype);
        add(m_col_gndelevmin);
        add(m_col_gndelevmax);
        add(m_col_altlwr);
        add(m_col_altlwr_agl);
        add(m_col_altlwr_fl);
        add(m_col_altlwr_sfc);
        add(m_col_altlwr_gnd);
        add(m_col_altlwr_unlim);
        add(m_col_altlwr_notam);
        add(m_col_altlwr_unkn);
        add(m_col_altupr);
        add(m_col_altupr_agl);
        add(m_col_altupr_fl);
        add(m_col_altupr_sfc);
        add(m_col_altupr_gnd);
        add(m_col_altupr_unlim);
        add(m_col_altupr_notam);
        add(m_col_altupr_unkn);
        add(m_col_modtime);
}

AirspaceEditor::AirspaceSegmentModelColumns::AirspaceSegmentModelColumns( void )
{
        add(m_col_segment);
        add(m_col_lon1);
        add(m_col_lat1);
        add(m_col_lon2);
        add(m_col_lat2);
        add(m_col_lon0);
        add(m_col_lat0);
        add(m_col_shape);
        add(m_col_radius1);
        add(m_col_radius2);
}

AirspaceEditor::AirspacePolygonModelColumns::AirspacePolygonModelColumns( void )
{
        add(m_col_point);
        add(m_col_lon);
        add(m_col_lat);
}

AirspaceEditor::AsyncElevUpdate::AsyncElevUpdate(Engine& eng, AirspacesDb & db, const AirspacesDb::Address & addr, const sigc::slot<void,AirspacesDb::Airspace&>& save)
        : m_db(db), m_addr(addr), m_save(save), m_refcnt(2), m_asyncref(true)
{
        std::cerr << "AirspaceEditor::AsyncElevUpdate::AsyncElevUpdate: " << (long)this << std::endl;
        AirspacesDb::Airspace e(m_db(m_addr));
        if (!e.is_valid())
                return;
        m_dispatch.connect(sigc::mem_fun(*this, &AirspaceEditor::AsyncElevUpdate::callback));
        m_async = eng.async_elevation_minmax(e.get_polygon());
        if (m_async) {
                m_async->connect(sigc::mem_fun(*this, &AirspaceEditor::AsyncElevUpdate::callback_dispatch));
        } else {
                m_asyncref = false;
                unreference();
        }
}

AirspaceEditor::AsyncElevUpdate::~AsyncElevUpdate()
{
        cancel();
        std::cerr << "AirspaceEditor::AsyncElevUpdate::~AsyncElevUpdate: " << (long)this << std::endl;
}

void AirspaceEditor::AsyncElevUpdate::reference(void )
{
        ++m_refcnt;
}

void AirspaceEditor::AsyncElevUpdate::unreference(void )
{
        if (!(--m_refcnt))
                delete this;
}

void AirspaceEditor::AsyncElevUpdate::cancel(void )
{
        Engine::ElevationMinMaxResult::cancel(m_async);
        if (m_asyncref) {
                m_asyncref = false;
                unreference();
        }
}

bool AirspaceEditor::AsyncElevUpdate::is_done(void )
{
        return !m_async;
}

void AirspaceEditor::AsyncElevUpdate::callback_dispatch(void )
{
        m_dispatch();
}

void AirspaceEditor::AsyncElevUpdate::callback(void )
{
        if (!m_async)
                return;
        if (!m_async->is_done())
                return;
        if (!m_async->is_error()) {
                AirspacesDb::Airspace e(m_db(m_addr));
                if (e.is_valid()) {
                        e.set_modtime(time(0));
                        TopoDb30::minmax_elev_t elev(m_async->get_result());
                        std::cerr << "gnd elev: " << elev.first << " / " << elev.second << std::endl;
                        if (elev.first != TopoDb30::nodata)
                                e.set_gndelevmin(Point::round<AirspacesDb::Airspace::gndelev_t,float>(elev.first * Point::m_to_ft));
                        if (elev.second != TopoDb30::nodata)
                                e.set_gndelevmax(Point::round<AirspacesDb::Airspace::gndelev_t,float>(elev.second * Point::m_to_ft));
                        if (elev.first != TopoDb30::nodata || elev.second != TopoDb30::nodata)
                                m_save(e);
                }
        }
        cancel();
}

AirspaceEditor::AirspaceEditor(const std::string & dir_main, Engine::auxdb_mode_t auxdbmode, const std::string & dir_aux)
        : m_dbchanged(false), m_engine(dir_main, auxdbmode, dir_aux), m_airspaceeditorwindow(0), m_airspaceeditortreeview(0), m_airspaceeditorstatusbar(0),
          m_airspaceeditorfind(0), m_airspaceeditornotebook(0), m_airspaceeditorclassexcept(0), m_airspaceeditorefftime(0),
          m_airspaceeditornote(0), m_airspaceshapeeditortreeview(0), m_airspacepolyapproxeditortreeview(0),
          m_airspaceeditormap(0), m_airspaceeditorundo(0), m_airspaceeditorredo(0),
          m_airspaceeditornewairspace(0), m_airspaceeditorduplicateairspace(0), m_airspaceeditordeleteairspace(0), m_airspaceeditormapenable(0),  m_airspaceeditorairportdiagramenable(0),
          m_airspaceeditormapzoomin(0), m_airspaceeditormapzoomout(0),
          m_airspaceeditorappendshape(0), m_airspaceeditorinsertshape(0), m_airspaceeditordeleteshape(0),
          m_airspaceeditorselectall(0), m_airspaceeditorunselectall(0), m_aboutdialog(0),
          m_curentryid(0)
{
        m_db.open(dir_main);
        std::cerr << "Main database " << (dir_main.empty() ? std::string("(default)") : dir_main) << std::endl;
        {
                Glib::ustring dir_aux1(m_engine.get_aux_dir(auxdbmode, dir_aux));
                if (!dir_aux1.empty() && m_db.is_dbfile_exists(dir_aux1))
                        m_db.attach(dir_aux1);
                if (m_db.is_aux_attached())
                        std::cerr << "Auxillary airspaces database " << dir_aux1 << " attached" << std::endl;
        }
        m_refxml = load_glade_file("dbeditor.glade", "airporteditorwindow");
        m_refxml->get_widget_derived("airspaceeditorwindow", m_airspaceeditorwindow);
        m_refxml->get_widget("airspaceeditortreeview", m_airspaceeditortreeview);
        m_refxml->get_widget("airspaceeditorstatusbar", m_airspaceeditorstatusbar);
        m_refxml->get_widget("airspaceeditorfind", m_airspaceeditorfind);
        m_refxml->get_widget("airspaceeditornotebook", m_airspaceeditornotebook);
        m_refxml->get_widget("airspaceeditorclassexcept", m_airspaceeditorclassexcept);
        m_refxml->get_widget("airspaceeditorefftime", m_airspaceeditorefftime);
        m_refxml->get_widget("airspaceeditornote", m_airspaceeditornote);
        m_refxml->get_widget("airspaceshapeeditortreeview", m_airspaceshapeeditortreeview);
        m_refxml->get_widget("airspacepolyapproxeditortreeview", m_airspacepolyapproxeditortreeview);
        m_refxml->get_widget_derived("airspaceeditormap", m_airspaceeditormap);
        {
                Glib::RefPtr<Gnome::Glade::Xml> refxml = load_glade_file_nosearch(m_refxml->get_filename(), "aboutdialog");
                refxml->get_widget("aboutdialog", m_aboutdialog);
                m_aboutdialog->signal_response().connect(sigc::mem_fun(*this, &AirspaceEditor::aboutdialog_response));
        }
        Gtk::Button *buttonclear(0);
        m_refxml->get_widget("airspaceeditorclearbutton", buttonclear);
        buttonclear->signal_clicked().connect(sigc::mem_fun(*this, &AirspaceEditor::buttonclear_clicked));
        m_airspaceeditorfind->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::find_changed));
        Gtk::MenuItem *menu_quit(0), *menu_cut(0), *menu_copy(0), *menu_paste(0), *menu_delete(0);
        Gtk::MenuItem *menu_preferences(0), *menu_about(0);
        m_refxml->get_widget("airspaceeditorquit", menu_quit);
        m_refxml->get_widget("airspaceeditorundo", m_airspaceeditorundo);
        m_refxml->get_widget("airspaceeditorredo", m_airspaceeditorredo);
        m_refxml->get_widget("airspaceeditornewairspace", m_airspaceeditornewairspace);
        m_refxml->get_widget("airspaceeditorduplicateairspace", m_airspaceeditorduplicateairspace);
        m_refxml->get_widget("airspaceeditordeleteairspace", m_airspaceeditordeleteairspace);
        m_refxml->get_widget("airspaceeditorplacelabel", m_airspaceeditorplacelabel);
        m_refxml->get_widget("airspaceeditorgndelev", m_airspaceeditorgndelev);
        m_refxml->get_widget("airspaceeditormapenable", m_airspaceeditormapenable);
        m_refxml->get_widget("airspaceeditorairportdiagramenable", m_airspaceeditorairportdiagramenable);
        m_refxml->get_widget("airspaceeditormapzoomin", m_airspaceeditormapzoomin);
        m_refxml->get_widget("airspaceeditormapzoomout", m_airspaceeditormapzoomout);
        m_refxml->get_widget("airspaceeditorappendshape", m_airspaceeditorappendshape);
        m_refxml->get_widget("airspaceeditorinsertshape", m_airspaceeditorinsertshape);
        m_refxml->get_widget("airspaceeditordeleteshape", m_airspaceeditordeleteshape);
        m_refxml->get_widget("airspaceeditorcut", menu_cut);
        m_refxml->get_widget("airspaceeditorcopy", menu_copy);
        m_refxml->get_widget("airspaceeditorpaste", menu_paste);
        m_refxml->get_widget("airspaceeditordelete", menu_delete);
        m_refxml->get_widget("airspaceeditorselectall", m_airspaceeditorselectall);
        m_refxml->get_widget("airspaceeditorunselectall", m_airspaceeditorunselectall);
        m_refxml->get_widget("airspaceeditorpreferences", menu_preferences);
        m_refxml->get_widget("airspaceeditorabout", menu_about);
        hide_notebook_pages();
        menu_quit->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_quit_activate));
        m_airspaceeditorundo->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_undo_activate));
        m_airspaceeditorredo->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_redo_activate));
        m_airspaceeditornewairspace->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_newairspace_activate));
        m_airspaceeditorduplicateairspace->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_duplicateairspace_activate));
        m_airspaceeditordeleteairspace->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_deleteairspace_activate));
        m_airspaceeditorplacelabel->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_placelabel_activate));
        m_airspaceeditorgndelev->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_gndelev_activate));
        m_airspaceeditormapenable->signal_toggled().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_mapenable_toggled));
        m_airspaceeditorairportdiagramenable->signal_toggled().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_airportdiagramenable_toggled));
        m_airspaceeditormapzoomin->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_mapzoomin_activate));
        m_airspaceeditormapzoomout->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_mapzoomout_activate));
        if (m_airspaceeditormapenable->get_active()) {
                m_airspaceeditormapzoomin->show();
                m_airspaceeditormapzoomout->show();
        } else {
                m_airspaceeditormapzoomin->hide();
                m_airspaceeditormapzoomout->hide();
        }
        m_airspaceeditorwindow->signal_zoomin().connect(sigc::bind_return(sigc::mem_fun(*this, &AirspaceEditor::menu_mapzoomin_activate), true));
        m_airspaceeditorwindow->signal_zoomout().connect(sigc::bind_return(sigc::mem_fun(*this, &AirspaceEditor::menu_mapzoomout_activate), true));
        menu_cut->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_cut_activate));
        menu_copy->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_copy_activate));
        menu_paste->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_paste_activate));
        menu_delete->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_delete_activate));
        m_airspaceeditorselectall->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_selectall_activate));
        m_airspaceeditorunselectall->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_unselectall_activate));
        m_airspaceeditorappendshape->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_appendshape_activate));
        m_airspaceeditorinsertshape->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_insertshape_activate));
        m_airspaceeditordeleteshape->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_deleteshape_activate));
        menu_preferences->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_prefs_activate));
        menu_about->signal_activate().connect(sigc::mem_fun(*this, &AirspaceEditor::menu_about_activate));
        update_undoredo_menu();
        // Vector Map
        m_airspaceeditormap->set_engine(m_engine);
        m_airspaceeditormap->set_renderer(VectorMapArea::renderer_vmap);
        m_airspaceeditormap->signal_cursor().connect(sigc::mem_fun(*this, &AirspaceEditor::map_cursor));
        m_airspaceeditormap->set_map_scale(m_engine.get_prefs().mapscale);
        AirspaceEditor::map_drawflags(m_engine.get_prefs().mapflags);
        m_engine.get_prefs().mapflags.signal_change().connect(sigc::mem_fun(*this, &AirspaceEditor::map_drawflags));
        // Airspace List
        m_airspaceliststore = Gtk::ListStore::create(m_airspacelistcolumns);
        m_airspaceeditortreeview->set_model(m_airspaceliststore);
        Gtk::TreeViewColumn *icao_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("ICAO"), m_airspacelistcolumns.m_col_icao) - 1);
        if (icao_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(icao_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_icao));
                icao_column->set_sort_column(m_airspacelistcolumns.m_col_icao);
        }
        Gtk::TreeViewColumn *name_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Name"), m_airspacelistcolumns.m_col_name) - 1);
        if (name_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(name_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_name));
                name_column->set_sort_column(m_airspacelistcolumns.m_col_name);
        }
        Gtk::TreeViewColumn *ident_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Ident"), m_airspacelistcolumns.m_col_ident) - 1);
        if (ident_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(ident_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_ident));
                ident_column->set_sort_column(m_airspacelistcolumns.m_col_ident);
        }
        LatLonCellRenderer *labellon_renderer = Gtk::manage(new LatLonCellRenderer(false));
        Gtk::TreeViewColumn *labellon_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Label Lon"), *labellon_renderer) - 1);
        if (labellon_column) {
                labellon_column->add_attribute(*labellon_renderer, labellon_renderer->get_property_name(), m_airspacelistcolumns.m_col_labellon);
                labellon_column->set_sort_column(m_airspacelistcolumns.m_col_labellon);
        }
        labellon_renderer->set_property("editable", true);
        labellon_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_labellon));
        LatLonCellRenderer *labellat_renderer = Gtk::manage(new LatLonCellRenderer(true));
        Gtk::TreeViewColumn *labellat_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Label Lat"), *labellat_renderer) - 1);
        if (labellat_column) {
                labellat_column->add_attribute(*labellat_renderer, labellat_renderer->get_property_name(), m_airspacelistcolumns.m_col_labellat);
                labellat_column->set_sort_column(m_airspacelistcolumns.m_col_labellat);
        }
        labellat_renderer->set_property("editable", true);
        labellat_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_labellat));
        LabelPlacementRenderer *label_placement_renderer = Gtk::manage(new LabelPlacementRenderer());
        Gtk::TreeViewColumn *label_placement_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Label"), *label_placement_renderer) - 1);
        if (label_placement_column) {
                label_placement_column->add_attribute(*label_placement_renderer, label_placement_renderer->get_property_name(), m_airspacelistcolumns.m_col_label_placement);
                label_placement_column->set_sort_column(m_airspacelistcolumns.m_col_label_placement);
        }
        label_placement_renderer->set_property("editable", true);
        label_placement_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_label_placement));
        AirspaceClassRenderer *airspaceclass_renderer = Gtk::manage(new AirspaceClassRenderer());
        Gtk::TreeViewColumn *airspaceclass_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Class"), *airspaceclass_renderer) - 1);
        if (airspaceclass_column) {
                airspaceclass_column->add_attribute(*airspaceclass_renderer, airspaceclass_renderer->get_property_name(), m_airspacelistcolumns.m_col_classtype);
                airspaceclass_column->set_sort_column(m_airspacelistcolumns.m_col_classtype);
        }
        airspaceclass_renderer->set_property("editable", true);
        airspaceclass_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_airspaceclass));
        Gtk::TreeViewColumn *commname_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Comm"), m_airspacelistcolumns.m_col_commname) - 1);
        if (commname_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(commname_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_commname));
                commname_column->set_sort_column(m_airspacelistcolumns.m_col_commname);
        }
        FrequencyCellRenderer *commfreq0_renderer = Gtk::manage(new FrequencyCellRenderer());
        Gtk::TreeViewColumn *commfreq0_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Freq 0"), *commfreq0_renderer) - 1);
        if (commfreq0_column) {
                commfreq0_column->add_attribute(*commfreq0_renderer, commfreq0_renderer->get_property_name(), m_airspacelistcolumns.m_col_commfreq0);
                commfreq0_column->set_sort_column(m_airspacelistcolumns.m_col_commfreq0);
        }
        commfreq0_renderer->set_property("editable", true);
        commfreq0_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_commfreq0));
        commfreq0_renderer->set_property("xalign", 1.0);
        FrequencyCellRenderer *commfreq1_renderer = Gtk::manage(new FrequencyCellRenderer());
        Gtk::TreeViewColumn *commfreq1_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Freq 1"), *commfreq1_renderer) - 1);
        if (commfreq1_column) {
                commfreq1_column->add_attribute(*commfreq1_renderer, commfreq1_renderer->get_property_name(), m_airspacelistcolumns.m_col_commfreq1);
                commfreq1_column->set_sort_column(m_airspacelistcolumns.m_col_commfreq1);
        }
        commfreq1_renderer->set_property("editable", true);
        commfreq1_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_commfreq1));
        commfreq1_renderer->set_property("xalign", 1.0);
        Gtk::TreeViewColumn *altlwr_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Lower Altitude"), m_airspacelistcolumns.m_col_altlwr) - 1);
        if (altlwr_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(altlwr_column->get_first_cell_renderer()));
                if (renderer) {
                        renderer->set_property("xalign", 1.0);
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr));
                }
                altlwr_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr);
        }
        Gtk::TreeViewColumn *altlwr_agl_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("AGL"), m_airspacelistcolumns.m_col_altlwr_agl) - 1);
        if (altlwr_agl_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_agl_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_agl));
                altlwr_agl_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_agl);
        }
        Gtk::TreeViewColumn *altlwr_fl_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("FL"), m_airspacelistcolumns.m_col_altlwr_fl) - 1);
        if (altlwr_fl_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_fl_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_fl));
                altlwr_fl_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_fl);
        }
        Gtk::TreeViewColumn *altlwr_sfc_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("SFC"), m_airspacelistcolumns.m_col_altlwr_sfc) - 1);
        if (altlwr_sfc_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_sfc_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_sfc));
                altlwr_sfc_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_sfc);
        }
        Gtk::TreeViewColumn *altlwr_gnd_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("GND"), m_airspacelistcolumns.m_col_altlwr_gnd) - 1);
        if (altlwr_gnd_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_gnd_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_gnd));
                altlwr_gnd_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_gnd);
        }
        Gtk::TreeViewColumn *altlwr_unlim_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Unlim"), m_airspacelistcolumns.m_col_altlwr_unlim) - 1);
        if (altlwr_unlim_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_unlim_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_unlim));
                altlwr_unlim_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_unlim);
        }
        Gtk::TreeViewColumn *altlwr_notam_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("NOTAM"), m_airspacelistcolumns.m_col_altlwr_notam) - 1);
        if (altlwr_notam_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_notam_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_notam));
                altlwr_notam_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_notam);
        }
        Gtk::TreeViewColumn *altlwr_unkn_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Unknown"), m_airspacelistcolumns.m_col_altlwr_unkn) - 1);
        if (altlwr_unkn_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altlwr_unkn_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altlwr_flag), AirspacesDb::Airspace::altflag_unkn));
                altlwr_unkn_column->set_sort_column(m_airspacelistcolumns.m_col_altlwr_unkn);
        }
        Gtk::TreeViewColumn *altupr_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Upper Altitude"), m_airspacelistcolumns.m_col_altupr) - 1);
        if (altupr_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(altupr_column->get_first_cell_renderer()));
                if (renderer) {
                        renderer->set_property("xalign", 1.0);
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr));
                }
                altupr_column->set_sort_column(m_airspacelistcolumns.m_col_altupr);
        }
        Gtk::TreeViewColumn *altupr_agl_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("AGL"), m_airspacelistcolumns.m_col_altupr_agl) - 1);
        if (altupr_agl_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_agl_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_agl));
                altupr_agl_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_agl);
        }
        Gtk::TreeViewColumn *altupr_fl_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("FL"), m_airspacelistcolumns.m_col_altupr_fl) - 1);
        if (altupr_fl_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_fl_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_fl));
                altupr_fl_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_fl);
        }
        Gtk::TreeViewColumn *altupr_sfc_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("SFC"), m_airspacelistcolumns.m_col_altupr_sfc) - 1);
        if (altupr_sfc_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_sfc_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_sfc));
                altupr_sfc_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_sfc);
        }
        Gtk::TreeViewColumn *altupr_gnd_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("GND"), m_airspacelistcolumns.m_col_altupr_gnd) - 1);
        if (altupr_gnd_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_gnd_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_gnd));
                altupr_gnd_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_gnd);
        }
        Gtk::TreeViewColumn *altupr_unlim_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Unlim"), m_airspacelistcolumns.m_col_altupr_unlim) - 1);
        if (altupr_unlim_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_unlim_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_unlim));
                altupr_unlim_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_unlim);
        }
        Gtk::TreeViewColumn *altupr_notam_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("NOTAM"), m_airspacelistcolumns.m_col_altupr_notam) - 1);
        if (altupr_notam_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_notam_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_notam));
                altupr_notam_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_notam);
        }
        Gtk::TreeViewColumn *altupr_unkn_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Unknown"), m_airspacelistcolumns.m_col_altupr_unkn) - 1);
        if (altupr_unkn_column) {
                Gtk::CellRendererToggle *renderer(dynamic_cast<Gtk::CellRendererToggle *>(altupr_unkn_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &AirspaceEditor::edited_altupr_flag), AirspacesDb::Airspace::altflag_unkn));
                altupr_unkn_column->set_sort_column(m_airspacelistcolumns.m_col_altupr_unkn);
        }
        Gtk::TreeViewColumn *gndelevmin_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("GND Elev Min"), m_airspacelistcolumns.m_col_gndelevmin) - 1);
        if (gndelevmin_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(gndelevmin_column->get_first_cell_renderer()));
                if (renderer) {
                        renderer->set_property("xalign", 1.0);
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_gndelevmin));
                }
                gndelevmin_column->set_sort_column(m_airspacelistcolumns.m_col_gndelevmin);
        }
        Gtk::TreeViewColumn *gndelevmax_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("GND Elev Max"), m_airspacelistcolumns.m_col_gndelevmax) - 1);
        if (gndelevmax_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(gndelevmax_column->get_first_cell_renderer()));
                if (renderer) {
                        renderer->set_property("xalign", 1.0);
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_gndelevmax));
                }
                gndelevmax_column->set_sort_column(m_airspacelistcolumns.m_col_gndelevmax);
        }
        Gtk::TreeViewColumn *srcid_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column_editable(_("Source ID"), m_airspacelistcolumns.m_col_sourceid) - 1);
        if (srcid_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(srcid_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_sourceid));
                srcid_column->set_sort_column(m_airspacelistcolumns.m_col_sourceid);
        }
        DateTimeCellRenderer *modtime_renderer = Gtk::manage(new DateTimeCellRenderer());
        Gtk::TreeViewColumn *modtime_column = m_airspaceeditortreeview->get_column(m_airspaceeditortreeview->append_column(_("Mod Time"), *modtime_renderer) - 1);
        if (modtime_column) {
                modtime_column->add_attribute(*modtime_renderer, modtime_renderer->get_property_name(), m_airspacelistcolumns.m_col_modtime);
                modtime_column->set_sort_column(m_airspacelistcolumns.m_col_modtime);
        }
        modtime_renderer->set_property("editable", true);
        modtime_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_modtime));
        //m_airspaceeditortreeview->append_column_editable(_("Mod Time"), m_airspacelistcolumns.m_col_modtime);
        for (unsigned int i = 0; i < 30; ++i) {
                Gtk::TreeViewColumn *col = m_airspaceeditortreeview->get_column(i);
                if (!col)
                        continue;
                col->set_resizable(true);
                col->set_reorderable(true);
        }
        m_airspaceeditortreeview->columns_autosize();
        m_airspaceeditortreeview->set_enable_search(true);
        m_airspaceeditortreeview->set_search_column(m_airspacelistcolumns.m_col_name);
        // selection
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        airspaceeditor_selection->set_mode(Gtk::SELECTION_MULTIPLE);
        airspaceeditor_selection->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::airspaceeditor_selection_changed));
        //airspaceeditor_selection->set_select_function(sigc::mem_fun(*this, &AirspaceEditor::airspaceeditor_select_function));

        // steup notebook pages
        m_airspaceeditorclassexcept->get_buffer()->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_classexcept));
        m_airspaceeditorefftime->get_buffer()->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_efftime));
        m_airspaceeditornote->get_buffer()->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_note));
        m_airspacesegment_model = Gtk::ListStore::create(m_airspacesegment_model_columns);
        m_airspacepolygon_model = Gtk::ListStore::create(m_airspacepolygon_model_columns);
        m_airspaceshapeeditortreeview->set_model(m_airspacesegment_model);
        m_airspacepolyapproxeditortreeview->set_model(m_airspacepolygon_model);
        // Segment Treeview
        Gtk::TreeViewColumn *segmentnr_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Number"), m_airspacesegment_model_columns.m_col_segment) - 1);
        if (segmentnr_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(segmentnr_column->get_first_cell_renderer()));
                segmentnr_column->set_sort_column(m_airspacesegment_model_columns.m_col_segment);
        }
        AirspaceShapeRenderer *segmentshape_renderer = Gtk::manage(new AirspaceShapeRenderer());
        Gtk::TreeViewColumn *segmentshape_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Shape"), *segmentshape_renderer) - 1);
        if (segmentshape_column) {
                segmentshape_column->add_attribute(*segmentshape_renderer, segmentshape_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_shape);
                segmentshape_column->set_sort_column(m_airspacesegment_model_columns.m_col_shape);
        }
        segmentshape_renderer->set_property("editable", true);
        segmentshape_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_shape));
        LatLonCellRenderer *lon1_renderer = Gtk::manage(new LatLonCellRenderer(false));
        Gtk::TreeViewColumn *lon1_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lon 1"), *lon1_renderer) - 1);
        if (lon1_column) {
                lon1_column->add_attribute(*lon1_renderer, lon1_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lon1);
                lon1_column->set_sort_column(m_airspacesegment_model_columns.m_col_lon1);
        }
        lon1_renderer->set_property("editable", true);
        lon1_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lon1));
        LatLonCellRenderer *lat1_renderer = Gtk::manage(new LatLonCellRenderer(true));
        Gtk::TreeViewColumn *lat1_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lat 1"), *lat1_renderer) - 1);
        if (lat1_column) {
                lat1_column->add_attribute(*lat1_renderer, lat1_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lat1);
                lat1_column->set_sort_column(m_airspacesegment_model_columns.m_col_lat1);
        }
        lat1_renderer->set_property("editable", true);
        lat1_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lat1));
        LatLonCellRenderer *lon2_renderer = Gtk::manage(new LatLonCellRenderer(false));
        Gtk::TreeViewColumn *lon2_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lon 2"), *lon2_renderer) - 1);
        if (lon2_column) {
                lon2_column->add_attribute(*lon2_renderer, lon2_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lon2);
                lon2_column->set_sort_column(m_airspacesegment_model_columns.m_col_lon2);
        }
        lon2_renderer->set_property("editable", true);
        lon2_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lon2));
        LatLonCellRenderer *lat2_renderer = Gtk::manage(new LatLonCellRenderer(true));
        Gtk::TreeViewColumn *lat2_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lat 2"), *lat2_renderer) - 1);
        if (lat2_column) {
                lat2_column->add_attribute(*lat2_renderer, lat2_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lat2);
                lat2_column->set_sort_column(m_airspacesegment_model_columns.m_col_lat2);
        }
        lat2_renderer->set_property("editable", true);
        lat2_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lat2));
        LatLonCellRenderer *lon0_renderer = Gtk::manage(new LatLonCellRenderer(false));
        Gtk::TreeViewColumn *lon0_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lon 0"), *lon0_renderer) - 1);
        if (lon0_column) {
                lon0_column->add_attribute(*lon0_renderer, lon0_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lon0);
                lon0_column->set_sort_column(m_airspacesegment_model_columns.m_col_lon0);
        }
        lon0_renderer->set_property("editable", true);
        lon0_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lon0));
        LatLonCellRenderer *lat0_renderer = Gtk::manage(new LatLonCellRenderer(true));
        Gtk::TreeViewColumn *lat0_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column(_("Lat 0"), *lat0_renderer) - 1);
        if (lat0_column) {
                lat0_column->add_attribute(*lat0_renderer, lat0_renderer->get_property_name(), m_airspacesegment_model_columns.m_col_lat0);
                lat0_column->set_sort_column(m_airspacesegment_model_columns.m_col_lat0);
        }
        lat0_renderer->set_property("editable", true);
        lat0_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lat0));
        Gtk::TreeViewColumn *radius1_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column_editable(_("Radius 1"), m_airspacesegment_model_columns.m_col_radius1) - 1);
        if (radius1_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(radius1_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_radius1));
                radius1_column->set_sort_column(m_airspacesegment_model_columns.m_col_radius1);
        }
        Gtk::TreeViewColumn *radius2_column = m_airspaceshapeeditortreeview->get_column(m_airspaceshapeeditortreeview->append_column_editable(_("Radius 2"), m_airspacesegment_model_columns.m_col_radius2) - 1);
        if (radius2_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(radius2_column->get_first_cell_renderer()));
                if (renderer)
                        renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_radius2));
                radius2_column->set_sort_column(m_airspacesegment_model_columns.m_col_radius2);
        }
        // Polygon Treeview
        Gtk::TreeViewColumn *polynr_column = m_airspacepolyapproxeditortreeview->get_column(m_airspacepolyapproxeditortreeview->append_column(_("Number"), m_airspacepolygon_model_columns.m_col_point) - 1);
        if (polynr_column) {
                Gtk::CellRendererText *renderer(dynamic_cast<Gtk::CellRendererText *>(polynr_column->get_first_cell_renderer()));
                segmentnr_column->set_sort_column(m_airspacepolygon_model_columns.m_col_point);
        }
        LatLonCellRenderer *lon_renderer = Gtk::manage(new LatLonCellRenderer(false));
        Gtk::TreeViewColumn *lon_column = m_airspacepolyapproxeditortreeview->get_column(m_airspacepolyapproxeditortreeview->append_column(_("Lon"), *lon_renderer) - 1);
        if (lon_column) {
                lon_column->add_attribute(*lon_renderer, lon_renderer->get_property_name(), m_airspacepolygon_model_columns.m_col_lon);
                lon_column->set_sort_column(m_airspacepolygon_model_columns.m_col_lon);
        }
        lon_renderer->set_property("editable", false);
        //lon_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lon));
        LatLonCellRenderer *lat_renderer = Gtk::manage(new LatLonCellRenderer(true));
        Gtk::TreeViewColumn *lat_column = m_airspacepolyapproxeditortreeview->get_column(m_airspacepolyapproxeditortreeview->append_column(_("Lat"), *lat_renderer) - 1);
        if (lat_column) {
                lat_column->add_attribute(*lat_renderer, lat_renderer->get_property_name(), m_airspacepolygon_model_columns.m_col_lat);
                lat_column->set_sort_column(m_airspacepolygon_model_columns.m_col_lat);
        }
        lat_renderer->set_property("editable", false);
        //lat_renderer->signal_edited().connect(sigc::mem_fun(*this, &AirspaceEditor::edited_lat));
        // selection
        Glib::RefPtr<Gtk::TreeSelection> shape_selection = m_airspaceshapeeditortreeview->get_selection();
        shape_selection->set_mode(Gtk::SELECTION_MULTIPLE);
        shape_selection->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::airspaceeditor_shape_selection_changed));
        Glib::RefPtr<Gtk::TreeSelection> polygon_selection = m_airspacepolyapproxeditortreeview->get_selection();
        polygon_selection->set_mode(Gtk::SELECTION_NONE);
        polygon_selection->signal_changed().connect(sigc::mem_fun(*this, &AirspaceEditor::airspaceeditor_polygon_selection_changed));
        // notebook
        m_airspaceeditornotebook->signal_switch_page().connect(sigc::mem_fun(*this, &AirspaceEditor::notebook_switch_page));
        // final setup
        airspaceeditor_selection_changed();
        //set_airspacelist("");
        m_dbselectdialog.signal_byname().connect(sigc::mem_fun(*this, &AirspaceEditor::dbselect_byname));
        m_dbselectdialog.signal_byrectangle().connect(sigc::mem_fun(*this, &AirspaceEditor::dbselect_byrect));
        m_dbselectdialog.signal_bytime().connect(sigc::mem_fun(*this, &AirspaceEditor::dbselect_bytime));
        m_dbselectdialog.show();
        m_prefswindow = PrefsWindow::create(m_engine.get_prefs());
}

AirspaceEditor::~AirspaceEditor()
{
        if (m_dbchanged) {
                m_db.detach();
                std::cerr << "Analyzing database..." << std::endl;
                m_db.analyze();
                m_db.vacuum();
        }
        m_db.close();
}

void AirspaceEditor::dbselect_byname( const Glib::ustring & s, AirspacesDb::comp_t comp )
{
        AirspacesDb::elementvector_t els(m_db.find_by_text(s, 0, comp, 0));
        m_airspaceliststore->clear();
        for (AirspacesDb::elementvector_t::iterator ei(els.begin()), ee(els.end()); ei != ee; ++ei) {
                AirspacesDb::element_t& e(*ei);
                if (!e.is_valid())
                        continue;
                Gtk::TreeModel::Row row(*(m_airspaceliststore->append()));
                set_row(row, e);
        }
}

void AirspaceEditor::dbselect_byrect( const Rect & r )
{
        AirspacesDb::elementvector_t els(m_db.find_by_rect(r, 0));
        m_airspaceliststore->clear();
        for (AirspacesDb::elementvector_t::iterator ei(els.begin()), ee(els.end()); ei != ee; ++ei) {
                AirspacesDb::element_t& e(*ei);
                if (!e.is_valid())
                        continue;
                Gtk::TreeModel::Row row(*(m_airspaceliststore->append()));
                set_row(row, e);
        }
}

void AirspaceEditor::dbselect_bytime(time_t timefrom, time_t timeto)
{
        AirspacesDb::elementvector_t els(m_db.find_by_time(timefrom, timeto, 0));
        m_airspaceliststore->clear();
        for (AirspacesDb::elementvector_t::iterator ei(els.begin()), ee(els.end()); ei != ee; ++ei) {
                AirspacesDb::element_t& e(*ei);
                if (!e.is_valid())
                        continue;
                Gtk::TreeModel::Row row(*(m_airspaceliststore->append()));
                set_row(row, e);
        }
}

void AirspaceEditor::buttonclear_clicked( void )
{
        m_airspaceeditorfind->set_text("");
}

void AirspaceEditor::find_changed( void )
{
        dbselect_byname(m_airspaceeditorfind->get_text());
}

void AirspaceEditor::menu_quit_activate( void )
{
        m_airspaceeditorwindow->hide();
}

void AirspaceEditor::menu_undo_activate( void )
{
        {
                Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceshapeeditortreeview->get_selection();
                selection->unselect_all();
        }
        AirspacesDb::Airspace e(m_undoredo.undo());
        if (e.is_valid()) {
                std::cerr << "undo: " << e.get_nrsegments() << std::endl;
                save_nostack(e);
                select(e);
        }
        update_undoredo_menu();
}

void AirspaceEditor::menu_redo_activate( void )
{
        {
                Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceshapeeditortreeview->get_selection();
                selection->unselect_all();
        }
        AirspacesDb::Airspace e(m_undoredo.redo());
        if (e.is_valid()) {
                save_nostack(e);
                select(e);
        }
        update_undoredo_menu();
}

void AirspaceEditor::menu_newairspace_activate( void )
{
        AirspacesDb::Airspace e;
        e.set_bdryclass('A');
        e.set_typecode('A');
        e.set_sourceid(make_sourceid());
        e.set_modtime(time(0));
        save(e);
        select(e);
}

void AirspaceEditor::menu_duplicateairspace_activate(void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        if (sel.size() != 1)
                return;
        Gtk::TreeModel::iterator iter(m_airspaceliststore->get_iter(*sel.begin()));
        if (!iter) {
                std::cerr << "cannot get iterator for path " << (*sel.begin()).to_string() << " ??" << std::endl;
                return;
        }
        Gtk::TreeModel::Row row = *iter;
        std::cerr << "duplicating row: " << row[m_airspacelistcolumns.m_col_id] << std::endl;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        AirspacesDb::Airspace e2;
        e2.set_labelcoord(e.get_labelcoord());
        e2.set_icao(e.get_icao());
        e2.set_name(e.get_name());
        e2.set_ident(e.get_ident());
        e2.set_classexcept(e.get_classexcept());
        e2.set_efftime(e.get_efftime());
        e2.set_note(e.get_note());
        e2.set_commname(e.get_commname());
        e2.set_commfreq(0, e.get_commfreq(0));
        e2.set_commfreq(1, e.get_commfreq(1));
        e2.set_bdryclass(e.get_bdryclass());
        e2.set_typecode(e.get_typecode());
        e2.set_gndelevmin(e.get_gndelevmin());
        e2.set_gndelevmax(e.get_gndelevmax());
        e2.set_altlwr(e.get_altlwr());
        e2.set_altupr(e.get_altupr());
        e2.set_altlwrflags(e.get_altlwrflags());
        e2.set_altuprflags(e.get_altuprflags());
        e2.set_label_placement(e.get_label_placement());
        for (unsigned int i = 0; i < e.get_nrsegments(); i++)
                e2.add_segment(e[i]);
        e2.recompute_lineapprox();
        e2.recompute_bbox();
        e2.set_sourceid(make_sourceid());
        e2.set_modtime(time(0));
        save(e2);
        select(e2);
}

void AirspaceEditor::menu_deleteairspace_activate( void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        if (sel.empty())
                return;
        if (sel.size() > 1) {
                Gtk::MessageDialog *dlg = new Gtk::MessageDialog(_("Delete multiple airspaces?"), false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO, true);
                int res = dlg->run();
                delete dlg;
                if (res != Gtk::RESPONSE_YES) {
                        if (res != Gtk::RESPONSE_NO)
                                std::cerr << "unexpected dialog result: " << res << std::endl;
                        return;
                }
        }
        for (Gtk::TreeSelection::ListHandle_Path::iterator si(sel.begin()), se(sel.end()); si != se; si++) {
                Gtk::TreeModel::iterator iter(m_airspaceliststore->get_iter(*si));
                if (!iter) {
                        std::cerr << "cannot get iterator for path " << (*si).to_string() << " ??" << std::endl;
                        continue;
                }
                Gtk::TreeModel::Row row = *iter;
                std::cerr << "deleting row: " << row[m_airspacelistcolumns.m_col_id] << std::endl;
                AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
                m_undoredo.erase(e);
                m_db.erase(e);
                m_airspaceliststore->erase(iter);
        }
}

void AirspaceEditor::menu_placelabel_activate(void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        if (sel.empty())
                return;
        if (sel.size() > 1) {
                Gtk::MessageDialog *dlg = new Gtk::MessageDialog(_("Place labels for multiple airspaces?"), false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO, true);
                int res = dlg->run();
                delete dlg;
                if (res != Gtk::RESPONSE_YES) {
                        if (res != Gtk::RESPONSE_NO)
                                std::cerr << "unexpected dialog result: " << res << std::endl;
                        return;
                }
        }
        for (Gtk::TreeSelection::ListHandle_Path::iterator si(sel.begin()), se(sel.end()); si != se; si++) {
                Gtk::TreeModel::iterator iter(m_airspaceliststore->get_iter(*si));
                if (!iter) {
                        std::cerr << "cannot get iterator for path " << (*si).to_string() << " ??" << std::endl;
                        continue;
                }
                Gtk::TreeModel::Row row = *iter;
                std::cerr << "placing label in row: " << row[m_airspacelistcolumns.m_col_id] << std::endl;
                AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
                if (!e.is_valid())
                        continue;
                e.set_modtime(time(0));
                e.compute_initial_label_placement();
                save(e);
        }
}

void AirspaceEditor::menu_gndelev_activate(void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        if (sel.empty())
                return;
        if (sel.size() > 1) {
                Gtk::MessageDialog *dlg = new Gtk::MessageDialog(_("Recompute ground elevation for multiple airspaces?"), false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO, true);
                int res = dlg->run();
                delete dlg;
                if (res != Gtk::RESPONSE_YES) {
                        if (res != Gtk::RESPONSE_NO)
                                std::cerr << "unexpected dialog result: " << res << std::endl;
                        return;
                }
        }
        for (Gtk::TreeSelection::ListHandle_Path::iterator si(sel.begin()), se(sel.end()); si != se; si++) {
                Gtk::TreeModel::iterator iter(m_airspaceliststore->get_iter(*si));
                if (!iter) {
                        std::cerr << "cannot get iterator for path " << (*si).to_string() << " ??" << std::endl;
                        continue;
                }
                Gtk::TreeModel::Row row = *iter;
                std::cerr << "recomputing ground elevation for row: " << row[m_airspacelistcolumns.m_col_id] << std::endl;
                Glib::RefPtr<AsyncElevUpdate> aeu(new AsyncElevUpdate(m_engine, m_db, row[m_airspacelistcolumns.m_col_id], sigc::mem_fun(*this, &AirspaceEditor::save)));
        }
}

void AirspaceEditor::menu_selectall_activate(void)
{
        switch (m_airspaceeditornotebook->get_current_page()) {
                case 0:
                {
                        Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceeditortreeview->get_selection();
                        selection->select_all();
                        break;
                }

                case 2:
                {
                        Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceshapeeditortreeview->get_selection();
                        selection->select_all();
                        break;
                }
        }
}

void AirspaceEditor::menu_unselectall_activate(void)
{
        switch (m_airspaceeditornotebook->get_current_page()) {
                case 0:
                {
                        Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceeditortreeview->get_selection();
                        selection->unselect_all();
                        break;
                }

                case 2:
                {
                        Glib::RefPtr<Gtk::TreeSelection> selection = m_airspaceshapeeditortreeview->get_selection();
                        selection->unselect_all();
                        break;
                }
        }
}

void AirspaceEditor::menu_cut_activate( void )
{
        Gtk::Widget *focus = m_airspaceeditorwindow->get_focus();
        if (!focus)
                return;
        Gtk::Editable *edit = dynamic_cast<Gtk::Editable *>(focus);
        if (!edit)
                return;
        edit->cut_clipboard();
}

void AirspaceEditor::menu_copy_activate( void )
{
        Gtk::Widget *focus = m_airspaceeditorwindow->get_focus();
        if (!focus)
                return;
        Gtk::Editable *edit = dynamic_cast<Gtk::Editable *>(focus);
        if (!edit)
                return;
        edit->copy_clipboard();
}

void AirspaceEditor::menu_paste_activate( void )
{
        Gtk::Widget *focus = m_airspaceeditorwindow->get_focus();
        if (!focus)
                return;
        Gtk::Editable *edit = dynamic_cast<Gtk::Editable *>(focus);
        if (!edit)
                return;
        edit->paste_clipboard();
}

void AirspaceEditor::menu_delete_activate( void )
{
        Gtk::Widget *focus = m_airspaceeditorwindow->get_focus();
        if (!focus)
                return;
        Gtk::Editable *edit = dynamic_cast<Gtk::Editable *>(focus);
        if (!edit)
                return;
        edit->delete_selection();
}

void AirspaceEditor::menu_mapenable_toggled( void )
{
        if (m_airspaceeditormapenable->get_active()) {
                m_airspaceeditorairportdiagramenable->set_active(false);
                m_airspaceeditormap->set_renderer(VectorMapArea::renderer_vmap);
                m_airspaceeditormap->set_map_scale(m_engine.get_prefs().mapscale);
                m_airspaceeditormap->show();
                m_airspaceeditormapzoomin->show();
                m_airspaceeditormapzoomout->show();
        } else {
                m_airspaceeditormap->hide();
                m_airspaceeditormapzoomin->hide();
                m_airspaceeditormapzoomout->hide();
        }
}

void AirspaceEditor::menu_airportdiagramenable_toggled( void )
{
        if (m_airspaceeditorairportdiagramenable->get_active()) {
                m_airspaceeditormapenable->set_active(false);
                m_airspaceeditormap->set_renderer(VectorMapArea::renderer_airportdiagram);
                m_airspaceeditormap->set_map_scale(m_engine.get_prefs().mapscaleairportdiagram);
                m_airspaceeditormap->show();
                m_airspaceeditormapzoomin->show();
                m_airspaceeditormapzoomout->show();
        } else {
                m_airspaceeditormap->hide();
                m_airspaceeditormapzoomin->hide();
                m_airspaceeditormapzoomout->hide();
        }
}

void AirspaceEditor::menu_mapzoomin_activate( void )
{
        m_airspaceeditormap->zoom_in();
        if (m_airspaceeditormap->get_renderer() == VectorMapArea::renderer_airportdiagram)
                m_engine.get_prefs().mapscaleairportdiagram = m_airspaceeditormap->get_map_scale();
        else
                m_engine.get_prefs().mapscale = m_airspaceeditormap->get_map_scale();
}

void AirspaceEditor::menu_mapzoomout_activate( void )
{
        m_airspaceeditormap->zoom_out();
        if (m_airspaceeditormap->get_renderer() == VectorMapArea::renderer_airportdiagram)
                m_engine.get_prefs().mapscaleairportdiagram = m_airspaceeditormap->get_map_scale();
        else
                m_engine.get_prefs().mapscale = m_airspaceeditormap->get_map_scale();
}

void AirspaceEditor::renumber_shapes( void )
{
        unsigned int seg = 0;
        for (Gtk::TreeModel::iterator si(m_airspacesegment_model->children().begin()); si; si++, seg++) {
                Gtk::TreeRow row(*si);
                row[m_airspacesegment_model_columns.m_col_segment] = seg;
        }
}

void AirspaceEditor::new_shape(Gtk::TreeModel::iterator iter)
{
        Gtk::TreeModel::iterator iterp(iter), itern(iter);
        itern++;
        if (itern == m_airspacesegment_model->children().end())
                itern = m_airspacesegment_model->children().begin();
        if (iterp == m_airspacesegment_model->children().begin())
                iterp = m_airspacesegment_model->children().end();
        iterp--;
        Gtk::TreeRow row(*iter), rowp(*iterp), rown(*itern);
        row[m_airspacesegment_model_columns.m_col_shape] = '-';
        Point pt1(rowp[m_airspacesegment_model_columns.m_col_lon2], rowp[m_airspacesegment_model_columns.m_col_lat2]);
        Point pt2(rown[m_airspacesegment_model_columns.m_col_lon1], rown[m_airspacesegment_model_columns.m_col_lat1]);
        row[m_airspacesegment_model_columns.m_col_lat1] = pt1.get_lat();
        row[m_airspacesegment_model_columns.m_col_lon1] = pt1.get_lon();
        row[m_airspacesegment_model_columns.m_col_lat2] = pt2.get_lat();
        row[m_airspacesegment_model_columns.m_col_lon2] = pt2.get_lon();
        row[m_airspacesegment_model_columns.m_col_lat0] = 0;
        row[m_airspacesegment_model_columns.m_col_lon0] = 0;
        row[m_airspacesegment_model_columns.m_col_radius1] = 0;
        row[m_airspacesegment_model_columns.m_col_radius2] = 0;
        renumber_shapes();
}

void AirspaceEditor::menu_appendshape_activate( void )
{
        new_shape(m_airspacesegment_model->append());
        save_shapes();
}

void AirspaceEditor::menu_insertshape_activate( void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceshapeeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        Gtk::TreeSelection::ListHandle_Path::size_type selcount(sel.size());
        if (selcount != 1) {
                menu_appendshape_activate();
                return;
        }
        new_shape(m_airspacesegment_model->insert(m_airspacesegment_model->get_iter(*sel.begin())));
        save_shapes();
}

void AirspaceEditor::menu_deleteshape_activate( void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceshapeeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        for (Gtk::TreeSelection::ListHandle_Path::iterator si(sel.begin()), se(sel.end()); si != se; si++)
                m_airspacesegment_model->erase(m_airspacesegment_model->get_iter(*si));
        renumber_shapes();
        save_shapes();
}

void AirspaceEditor::menu_prefs_activate(void )
{
        if (m_prefswindow)
                m_prefswindow->show();
}

void AirspaceEditor::menu_about_activate( void )
{
        m_aboutdialog->show();
}

void AirspaceEditor::set_row_altlwr_flags( Gtk::TreeModel::Row & row, const AirspacesDb::Airspace & e )
{
        row[m_airspacelistcolumns.m_col_altlwr_agl] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_agl);
        row[m_airspacelistcolumns.m_col_altlwr_fl] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_fl);
        row[m_airspacelistcolumns.m_col_altlwr_sfc] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_sfc);
        row[m_airspacelistcolumns.m_col_altlwr_gnd] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_gnd);
        row[m_airspacelistcolumns.m_col_altlwr_unlim] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_unlim);
        row[m_airspacelistcolumns.m_col_altlwr_notam] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_notam);
        row[m_airspacelistcolumns.m_col_altlwr_unkn] = !!(e.get_altlwrflags() & AirspacesDb::Airspace::altflag_unkn);
}

void AirspaceEditor::set_row_altupr_flags( Gtk::TreeModel::Row & row, const AirspacesDb::Airspace & e )
{
        row[m_airspacelistcolumns.m_col_altupr_agl] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_agl);
        row[m_airspacelistcolumns.m_col_altupr_fl] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_fl);
        row[m_airspacelistcolumns.m_col_altupr_sfc] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_sfc);
        row[m_airspacelistcolumns.m_col_altupr_gnd] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_gnd);
        row[m_airspacelistcolumns.m_col_altupr_unlim] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_unlim);
        row[m_airspacelistcolumns.m_col_altupr_notam] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_notam);
        row[m_airspacelistcolumns.m_col_altupr_unkn] = !!(e.get_altuprflags() & AirspacesDb::Airspace::altflag_unkn);
}

void AirspaceEditor::set_row( Gtk::TreeModel::Row& row, const AirspacesDb::Airspace& e )
{
        row[m_airspacelistcolumns.m_col_id] = e.get_address();
        row[m_airspacelistcolumns.m_col_labellon] = e.get_labelcoord().get_lon();
        row[m_airspacelistcolumns.m_col_labellat] = e.get_labelcoord().get_lat();
        row[m_airspacelistcolumns.m_col_label_placement] = e.get_label_placement();
        row[m_airspacelistcolumns.m_col_icao] = e.get_icao();
        row[m_airspacelistcolumns.m_col_name] = e.get_name();
        row[m_airspacelistcolumns.m_col_ident] = e.get_ident();
        row[m_airspacelistcolumns.m_col_commname] = e.get_commname();
        row[m_airspacelistcolumns.m_col_commfreq0] = e.get_commfreq(0);
        row[m_airspacelistcolumns.m_col_commfreq1] = e.get_commfreq(1);
        row[m_airspacelistcolumns.m_col_classtype] = AirspaceClassType(e.get_bdryclass(), e.get_typecode());
        row[m_airspacelistcolumns.m_col_gndelevmin] = e.get_gndelevmin();
        row[m_airspacelistcolumns.m_col_gndelevmax] = e.get_gndelevmax();
        row[m_airspacelistcolumns.m_col_altlwr] = e.get_altlwr();
        set_row_altlwr_flags(row, e);
        row[m_airspacelistcolumns.m_col_altupr] = e.get_altupr();
        set_row_altupr_flags(row, e);
        row[m_airspacelistcolumns.m_col_sourceid] = e.get_sourceid();
        row[m_airspacelistcolumns.m_col_modtime] = e.get_modtime();
}

void AirspaceEditor::edited_labellon( const Glib::ustring & path_string, const Glib::ustring& new_text, Point::coord_t new_coord )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_labellon] = new_coord;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        Point pt(e.get_labelcoord());
        pt.set_lon(new_coord);
        e.set_labelcoord(pt);
        save_noerr(e, row);
        m_airspaceeditormap->set_center(pt, e.get_altlwr_corr());
        m_airspaceeditormap->set_cursor(pt);
}

void AirspaceEditor::edited_labellat( const Glib::ustring & path_string, const Glib::ustring& new_text, Point::coord_t new_coord )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_labellat] = new_coord;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        Point pt(e.get_labelcoord());
        pt.set_lat(new_coord);
        e.set_labelcoord(pt);
        save_noerr(e, row);
        m_airspaceeditormap->set_center(pt, e.get_altlwr_corr());
        m_airspaceeditormap->set_cursor(pt);
}

void AirspaceEditor::edited_gndelevmin( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        int32_t elev = Conversions::signed_feet_parse(new_text);
        row[m_airspacelistcolumns.m_col_gndelevmin] = elev;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_gndelevmin(elev);
        save_noerr(e, row);
}

void AirspaceEditor::edited_gndelevmax( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        int32_t elev = Conversions::signed_feet_parse(new_text);
        row[m_airspacelistcolumns.m_col_gndelevmax] = elev;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_gndelevmax(elev);
        save_noerr(e, row);
}

void AirspaceEditor::edited_altlwr( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        int32_t elev = Conversions::signed_feet_parse(new_text);
        row[m_airspacelistcolumns.m_col_altlwr] = elev;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_altlwr(elev);
        save_noerr(e, row);
}

void AirspaceEditor::edited_altupr( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        int32_t elev = Conversions::signed_feet_parse(new_text);
        row[m_airspacelistcolumns.m_col_altupr] = elev;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_altupr(elev);
        save_noerr(e, row);
}

void AirspaceEditor::edited_commname( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_commname] = new_text;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_commname(new_text);
        save_noerr(e, row);
}

void AirspaceEditor::edited_commfreq0( const Glib::ustring & path_string, const Glib::ustring & new_text, uint32_t new_freq )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_commfreq0] = new_freq;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_commfreq(0, new_freq);
        save_noerr(e, row);
}

void AirspaceEditor::edited_commfreq1( const Glib::ustring & path_string, const Glib::ustring & new_text, uint32_t new_freq )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_commfreq0] = new_freq;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_commfreq(1, new_freq);
        save_noerr(e, row);
}

void AirspaceEditor::edited_sourceid( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_sourceid] = new_text;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_sourceid(new_text);
        save_noerr(e, row);
}

void AirspaceEditor::edited_icao( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_icao] = new_text;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_icao(new_text);
        save_noerr(e, row);
}

void AirspaceEditor::edited_name( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_name] = new_text;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_name(new_text);
        save_noerr(e, row);
}

void AirspaceEditor::edited_ident( const Glib::ustring & path_string, const Glib::ustring & new_text )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_ident] = new_text;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_ident(new_text);
        save_noerr(e, row);
}

void AirspaceEditor::edited_airspaceclass( const Glib::ustring & path_string, const Glib::ustring & new_text, AirspaceClassType new_clstype )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        if (!new_clstype.get_type())
                return;
        row[m_airspacelistcolumns.m_col_classtype] = new_clstype;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_bdryclass(new_clstype.get_class());
        e.set_typecode(new_clstype.get_type());
        save_noerr(e, row);
}

void AirspaceEditor::edited_label_placement( const Glib::ustring & path_string, const Glib::ustring & new_text, NavaidsDb::Navaid::label_placement_t new_placement )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        //std::cerr << "edited_label_placement: txt " << new_text << " pl " << (int)new_placement << std::endl;
        row[m_airspacelistcolumns.m_col_label_placement] = new_placement;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_label_placement(new_placement);
        save_noerr(e, row);
}

void AirspaceEditor::edited_altlwr_flag( const Glib::ustring & path_string, uint8_t f)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_altlwrflags(e.get_altlwrflags() ^ f);
        set_row_altlwr_flags(row, e);
        save_noerr(e, row);
}

void AirspaceEditor::edited_altupr_flag( const Glib::ustring & path_string, uint8_t f)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_altuprflags(e.get_altuprflags() ^ f);
        set_row_altupr_flags(row, e);
        save_noerr(e, row);
}

void AirspaceEditor::edited_modtime( const Glib::ustring & path_string, const Glib::ustring& new_text, time_t new_time )
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspaceliststore->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_modtime] = new_time;
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(new_time);
        save_noerr(e, row);
}

void AirspaceEditor::edited_classexcept(void)
{
        if (!m_curentryid)
                return;
        Glib::ustring new_text(m_airspaceeditorclassexcept->get_buffer()->get_text());
        AirspacesDb::Airspace e(m_db(m_curentryid));
        if (!e.is_valid())
                return;
        if (new_text == e.get_classexcept())
                return;
        e.set_modtime(time(0));
        e.set_classexcept(new_text);
        save_noerr_lists(e);
}

void AirspaceEditor::edited_efftime(void)
{
        if (!m_curentryid)
                return;
        Glib::ustring new_text(m_airspaceeditorefftime->get_buffer()->get_text());
        AirspacesDb::Airspace e(m_db(m_curentryid));
        if (!e.is_valid())
                return;
        if (new_text == e.get_efftime())
                return;
        e.set_modtime(time(0));
        e.set_efftime(new_text);
        save_noerr_lists(e);
}

void AirspaceEditor::edited_note(void)
{
        if (!m_curentryid)
                return;
        Glib::ustring new_text(m_airspaceeditornote->get_buffer()->get_text());
        AirspacesDb::Airspace e(m_db(m_curentryid));
        if (!e.is_valid())
                return;
        if (new_text == e.get_note())
                return;
        e.set_modtime(time(0));
        e.set_note(new_text);
        save_noerr_lists(e);
}

void AirspaceEditor::update_radius12(const Gtk::TreeModel::iterator & iter)
{
        Gtk::TreeRow row = *iter;
        char shape(row[m_airspacesegment_model_columns.m_col_shape]);
        switch (shape) {
                case 'C':
                case 'A':
                        break;

                default:
                        row[m_airspacesegment_model_columns.m_col_radius1] = 0;
                        row[m_airspacesegment_model_columns.m_col_radius2] = 0;
                        break;

                case 'L':
                case 'R':
                        Point pt1(row[m_airspacesegment_model_columns.m_col_lon1], row[m_airspacesegment_model_columns.m_col_lat1]);
                        Point pt2(row[m_airspacesegment_model_columns.m_col_lon2], row[m_airspacesegment_model_columns.m_col_lat2]);
                        Point pt0(row[m_airspacesegment_model_columns.m_col_lon0], row[m_airspacesegment_model_columns.m_col_lat0]);
                        row[m_airspacesegment_model_columns.m_col_radius1] = pt0.spheric_distance_nmi(pt1) * (Point::from_deg / 60.0);
                        row[m_airspacesegment_model_columns.m_col_radius2] = pt0.spheric_distance_nmi(pt2) * (Point::from_deg / 60.0);
                        break;
        }
}

void AirspaceEditor::edited_shape(const Glib::ustring& path_string, const Glib::ustring& new_text, char new_shape)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_shape] = new_shape;
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lon1(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lon1] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon1], row[m_airspacesegment_model_columns.m_col_lat1]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        Gtk::TreeModel::iterator iterp(iter);
        if (iterp == m_airspacesegment_model->children().begin())
                iterp = m_airspacesegment_model->children().end();
        iterp--;
        Gtk::TreeRow rowp = *iterp;
        rowp[m_airspacesegment_model_columns.m_col_lon2] = pt.get_lon();
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lat1(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lat1] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon1], row[m_airspacesegment_model_columns.m_col_lat1]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        Gtk::TreeModel::iterator iterp(iter);
        if (iterp == m_airspacesegment_model->children().begin())
                iterp = m_airspacesegment_model->children().end();
        iterp--;
        Gtk::TreeRow rowp = *iterp;
        rowp[m_airspacesegment_model_columns.m_col_lat2] = pt.get_lat();
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lon2(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lon2] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon2], row[m_airspacesegment_model_columns.m_col_lat2]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        Gtk::TreeModel::iterator itern(iter);
        itern++;
        if (itern == m_airspacesegment_model->children().end())
                itern = m_airspacesegment_model->children().begin();
        Gtk::TreeRow rown = *itern;
        rown[m_airspacesegment_model_columns.m_col_lon1] = pt.get_lon();
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lat2(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lat2] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon2], row[m_airspacesegment_model_columns.m_col_lat2]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        Gtk::TreeModel::iterator itern(iter);
        itern++;
        if (itern == m_airspacesegment_model->children().end())
                itern = m_airspacesegment_model->children().begin();
        Gtk::TreeRow rown = *itern;
        rown[m_airspacesegment_model_columns.m_col_lat1] = pt.get_lat();
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lon0(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lon0] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon0], row[m_airspacesegment_model_columns.m_col_lat0]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_lat0(const Glib::ustring& path_string, const Glib::ustring& new_text, Point::coord_t new_coord)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_lat0] = new_coord;
        Point pt(row[m_airspacesegment_model_columns.m_col_lon0], row[m_airspacesegment_model_columns.m_col_lat0]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        update_radius12(iter);
        save_shapes();
}

void AirspaceEditor::edited_radius1(const Glib::ustring& path_string, const Glib::ustring& new_text)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_radius1] = strtod(new_text.c_str(), 0);
        save_shapes();
}

void AirspaceEditor::edited_radius2(const Glib::ustring& path_string, const Glib::ustring& new_text)
{
        Gtk::TreePath path(path_string);
        Gtk::TreeModel::iterator iter = m_airspacesegment_model->get_iter(path);
        if (!iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacesegment_model_columns.m_col_radius2] = strtod(new_text.c_str(), 0);
        save_shapes();
}

void AirspaceEditor::map_cursor( Point cursor, VectorMapArea::cursor_validity_t valid )
{
        if (valid != VectorMapArea::cursor_mouse)
                return;
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        Gtk::TreeSelection::ListHandle_Path::size_type selcount(sel.size());
        Gtk::TreeModel::iterator iter;
        if (selcount > 0)
                iter = m_airspaceliststore->get_iter(*sel.begin());
        if (selcount != 1 || !iter)
                return;
        Gtk::TreeRow row = *iter;
        row[m_airspacelistcolumns.m_col_labellon] = cursor.get_lon();
        row[m_airspacelistcolumns.m_col_labellat] = cursor.get_lat();
        AirspacesDb::Airspace e(m_db(row[m_airspacelistcolumns.m_col_id]));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.set_labelcoord(cursor);
        save_noerr(e, row);
}

void AirspaceEditor::map_drawflags(int flags)
{
        *m_airspaceeditormap = (MapRenderer::DrawFlags)flags;
}

void AirspaceEditor::save_nostack( AirspacesDb::Airspace & e )
{
        std::cerr << "save_nostack: " << e.get_nrsegments() << std::endl;
        if (!e.is_valid())
                return;
        AirspacesDb::Airspace::Address addr(e.get_address());
        m_db.save(e);
        m_dbchanged = true;
        if (m_curentryid == addr)
                m_curentryid = e.get_address();
        m_airspaceeditormap->airspaces_changed();
        Gtk::TreeIter iter = m_airspaceliststore->children().begin();
        while (iter) {
                Gtk::TreeRow row = *iter;
                if (addr == row[m_airspacelistcolumns.m_col_id]) {
                        set_row(row, e);
                        return;
                }
                iter++;
        }
        Gtk::TreeModel::Row row(*(m_airspaceliststore->append()));
        set_row(row, e);
}

void AirspaceEditor::save( AirspacesDb::Airspace & e )
{
        std::cerr << "save" << std::endl;
        save_nostack(e);
        m_undoredo.save(e);
        update_undoredo_menu();
}

void AirspaceEditor::save_noerr(AirspacesDb::Airspace & e, Gtk::TreeModel::Row & row)
{
        if (e.get_address() != row[m_airspacelistcolumns.m_col_id])
                throw std::runtime_error("AirportEditor::save: ID mismatch");
        try {
                save(e);
        } catch (const sqlite3x::database_error& dberr) {
                AirspacesDb::Airspace eold(m_db(row[m_airspacelistcolumns.m_col_id]));
                set_row(row, eold);
        }
}

void AirspaceEditor::select( const AirspacesDb::Airspace & e )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeIter iter = m_airspaceliststore->children().begin();
        while (iter) {
                Gtk::TreeRow row = *iter;
                if (e.get_address() == row[m_airspacelistcolumns.m_col_id]) {
                        airspaceeditor_selection->unselect_all();
                        airspaceeditor_selection->select(iter);
                        m_airspaceeditortreeview->scroll_to_row(Gtk::TreeModel::Path(iter));
                        return;
                }
                iter++;
        }
}

void AirspaceEditor::update_undoredo_menu( void )
{
        m_airspaceeditorundo->set_sensitive(m_undoredo.can_undo());
        m_airspaceeditorredo->set_sensitive(m_undoredo.can_redo());
}

void AirspaceEditor::col_latlon1_clicked(void)
{
        std::cerr << "col_latlon1_clicked" << std::endl;
}

void AirspaceEditor::notebook_switch_page(GtkNotebookPage* page, guint page_num)
{
        std::cerr << "notebook page " << page_num << std::endl;
        m_airspaceeditornewairspace->set_sensitive(!page_num);
        airspaceeditor_selection_changed();
        airspaceeditor_shape_selection_changed();
}

void AirspaceEditor::hide_notebook_pages(void)
{
        for (int pg = 1;; ++pg) {
                Gtk::Widget *w(m_airspaceeditornotebook->get_nth_page(pg));
                if (!w)
                        break;
                //w->hide();
                w->set_sensitive(false);
        }
}

void AirspaceEditor::show_notebook_pages(void)
{
        for (int pg = 1;; ++pg) {
                Gtk::Widget *w(m_airspaceeditornotebook->get_nth_page(pg));
                if (!w)
                        break;
                //w->show();
                w->set_sensitive(true);
        }
}

void AirspaceEditor::set_segmentpolygonlist(AirspacesDb::Airspace::Address addr)
{
        m_curentryid = 0;
        m_airspacesegment_model->clear();
        m_airspacepolygon_model->clear();
        if (!addr) {
          unselected_out:
                m_airspaceeditorclassexcept->get_buffer()->set_text("");
                m_airspaceeditorefftime->get_buffer()->set_text("");
                m_airspaceeditornote->get_buffer()->set_text("");
                airspaceeditor_shape_selection_changed();
                return;
        }
        AirspacesDb::Airspace e = m_db(addr);
        if (!e.is_valid()) {
                goto unselected_out;
        }
        std::cerr << "Name: " << e.get_name() << " BBox: "
                  << e.get_bbox().get_southwest().get_lon_str2() << "," << e.get_bbox().get_southwest().get_lat_str2() << " -> "
                  << e.get_bbox().get_northeast().get_lon_str2() << "," << e.get_bbox().get_northeast().get_lat_str2() << std::endl;
        Gtk::TreeModel::iterator iter(m_airspacesegment_model->children().begin());
        for (unsigned int s = 0; s < e.get_nrsegments(); s++) {
                const AirspacesDb::Airspace::Segment& seg(e[s]);
                Gtk::TreeModel::Row row;
                if (!iter)
                        iter = m_airspacesegment_model->append();
                row = *iter;
                iter++;
                row[m_airspacesegment_model_columns.m_col_segment] = s;
                row[m_airspacesegment_model_columns.m_col_shape] = seg.get_shape();
                row[m_airspacesegment_model_columns.m_col_lat1] = seg.get_coord1().get_lat();
                row[m_airspacesegment_model_columns.m_col_lon1] = seg.get_coord1().get_lon();
                row[m_airspacesegment_model_columns.m_col_lat2] = seg.get_coord2().get_lat();
                row[m_airspacesegment_model_columns.m_col_lon2] = seg.get_coord2().get_lon();
                row[m_airspacesegment_model_columns.m_col_lat0] = seg.get_coord0().get_lat();
                row[m_airspacesegment_model_columns.m_col_lon0] = seg.get_coord0().get_lon();
                row[m_airspacesegment_model_columns.m_col_radius1] = seg.get_radius1() * (Point::to_deg * 60.0);
                row[m_airspacesegment_model_columns.m_col_radius2] = seg.get_radius2() * (Point::to_deg * 60.0);
        }
        for (; iter; ) {
                Gtk::TreeModel::iterator iter1(iter);
                iter++;
                m_airspacesegment_model->erase(iter1);
        }
        {
                const Polygon& poly(e.get_polygon());
                for (unsigned int p = 0; p < poly.size(); ++p) {
                        Gtk::TreeModel::Row row(*(m_airspacepolygon_model->append()));
                        row[m_airspacepolygon_model_columns.m_col_point] = p;
                        row[m_airspacepolygon_model_columns.m_col_lat] = poly[p].get_lat();
                        row[m_airspacepolygon_model_columns.m_col_lon] = poly[p].get_lon();
                }
        }
        m_airspaceeditorclassexcept->get_buffer()->set_text(e.get_classexcept());
        m_airspaceeditorefftime->get_buffer()->set_text(e.get_efftime());
        m_airspaceeditornote->get_buffer()->set_text(e.get_note());
        m_curentryid = addr;
}

void AirspaceEditor::airspaceeditor_selection_changed( void )
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        Gtk::TreeSelection::ListHandle_Path::size_type selcount(sel.size());
        Gtk::TreeModel::iterator iter;
        if (selcount > 0)
                iter = m_airspaceliststore->get_iter(*sel.begin());
        if (selcount > 1) {
                // multiple selections
                std::cerr << "multiple selection" << std::endl;
                bool firstpage(!m_airspaceeditornotebook->get_current_page());
                m_airspaceeditorduplicateairspace->set_sensitive(false);
                m_airspaceeditordeleteairspace->set_sensitive(firstpage);
                m_airspaceeditorplacelabel->set_sensitive(firstpage);
                m_airspaceeditorgndelev->set_sensitive(firstpage);
                m_airspaceeditormap->invalidate_cursor();
                set_segmentpolygonlist(0);
                hide_notebook_pages();
                return;
        }
        if (!selcount || !iter) {
                std::cerr << "selection cleared" << std::endl;
                m_airspaceeditorduplicateairspace->set_sensitive(false);
                m_airspaceeditordeleteairspace->set_sensitive(false);
                m_airspaceeditorplacelabel->set_sensitive(false);
                m_airspaceeditorgndelev->set_sensitive(false);
                m_airspaceeditormap->invalidate_cursor();
                set_segmentpolygonlist(0);
                hide_notebook_pages();
                return;
        }
        Gtk::TreeModel::Row row = *iter;
        //Do something with the row.
        std::cerr << "selected row: " << row[m_airspacelistcolumns.m_col_id] << std::endl;
        bool firstpage(!m_airspaceeditornotebook->get_current_page());
        m_airspaceeditorduplicateairspace->set_sensitive(firstpage);
        m_airspaceeditordeleteairspace->set_sensitive(firstpage);
        m_airspaceeditorplacelabel->set_sensitive(firstpage);
        m_airspaceeditorgndelev->set_sensitive(firstpage);
        Point pt(row[m_airspacelistcolumns.m_col_labellon], row[m_airspacelistcolumns.m_col_labellat]);
        m_airspaceeditormap->set_center(pt, 0);
        m_airspaceeditormap->set_cursor(pt);
        set_segmentpolygonlist(row[m_airspacelistcolumns.m_col_id]);
        show_notebook_pages();
}

void AirspaceEditor::save_shapes(void)
{
        if (!m_curentryid)
                return;
        AirspacesDb::Airspace e(m_db(m_curentryid));
        if (!e.is_valid())
                return;
        e.set_modtime(time(0));
        e.clear_segments();
        for (Gtk::TreeModel::iterator si(m_airspacesegment_model->children().begin()); si; si++) {
                Gtk::TreeRow row(*si);
                AirspacesDb::Airspace::Segment seg(Point(row[m_airspacesegment_model_columns.m_col_lon1], row[m_airspacesegment_model_columns.m_col_lat1]),
                                                   Point(row[m_airspacesegment_model_columns.m_col_lon2], row[m_airspacesegment_model_columns.m_col_lat2]),
                                                   Point(row[m_airspacesegment_model_columns.m_col_lon0], row[m_airspacesegment_model_columns.m_col_lat0]),
                                                   row[m_airspacesegment_model_columns.m_col_shape],
                                                   Point::round<int32_t,double>(row[m_airspacesegment_model_columns.m_col_radius1] * (Point::from_deg / 60.0)),
                                                   Point::round<int32_t,double>(row[m_airspacesegment_model_columns.m_col_radius2] * (Point::from_deg / 60.0)));
                e.add_segment(seg);
        }
        e.recompute_lineapprox();
        save_noerr_lists(e);
}

void AirspaceEditor::save_noerr_lists(AirspacesDb::Airspace & e)
{
        if (!m_curentryid)
                return;
        try {
                save(e);
        } catch (const sqlite3x::database_error& dberr) {
                e = m_db(m_curentryid);
        }
        set_segmentpolygonlist(e.get_address());
}

void AirspaceEditor::airspaceeditor_shape_selection_changed(void)
{
        if (!m_curentryid || 2 != m_airspaceeditornotebook->get_current_page()) {
                m_airspaceeditorappendshape->set_sensitive(false);
                m_airspaceeditorinsertshape->set_sensitive(false);
                m_airspaceeditordeleteshape->set_sensitive(false);
                return;
        }
        m_airspaceeditorappendshape->set_sensitive(true);
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspaceshapeeditortreeview->get_selection();
        Gtk::TreeSelection::ListHandle_Path sel(airspaceeditor_selection->get_selected_rows());
        Gtk::TreeSelection::ListHandle_Path::size_type selcount(sel.size());
        Gtk::TreeModel::iterator iter;
        if (selcount > 0)
                iter = m_airspacesegment_model->get_iter(*sel.begin());
        if (selcount > 1) {
                // multiple selections
                std::cerr << "multiple selection" << std::endl;
                m_airspaceeditorinsertshape->set_sensitive(false);
                m_airspaceeditordeleteshape->set_sensitive(true);
                return;
        }
        if (!selcount || !iter) {
                std::cerr << "selection cleared" << std::endl;
                m_airspaceeditorinsertshape->set_sensitive(false);
                m_airspaceeditordeleteshape->set_sensitive(false);
                return;
        }
        Gtk::TreeModel::Row row = *iter;
        //Do something with the row.
        std::cerr << "selected row: " << row[m_airspacesegment_model_columns.m_col_segment] << std::endl;
        m_airspaceeditorinsertshape->set_sensitive(true);
        m_airspaceeditordeleteshape->set_sensitive(true);
}

void AirspaceEditor::airspaceeditor_polygon_selection_changed(void)
{
        Glib::RefPtr<Gtk::TreeSelection> airspaceeditor_selection = m_airspacepolyapproxeditortreeview->get_selection();
        Gtk::TreeModel::iterator iter = airspaceeditor_selection->get_selected();
        if (iter) {
                Gtk::TreeModel::Row row = *iter;
                //Do something with the row.
                std::cerr << "selected row: " << row[m_airspacepolygon_model_columns.m_col_point] << std::endl;
                return;
        }
        std::cerr << "selection cleared" << std::endl;
}

void AirspaceEditor::aboutdialog_response( int response )
{
        if (response == Gtk::RESPONSE_CLOSE || response == Gtk::RESPONSE_CANCEL)
                m_aboutdialog->hide();
}

int main(int argc, char *argv[])
{
        try {
                std::string dir_main(""), dir_aux("");
                Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs;
                bool dis_aux(false);
                Glib::init();
                Glib::OptionGroup optgroup("airspaceeditor", "Airspace Editor Options", "Options controlling the Airspace Editor");
                Glib::OptionEntry optmaindir, optauxdir, optdisableaux;
                optmaindir.set_short_name('m');
                optmaindir.set_long_name("maindir");
                optmaindir.set_description("Directory containing the main database");
                optmaindir.set_arg_description("DIR");
                optauxdir.set_short_name('a');
                optauxdir.set_long_name("auxdir");
                optauxdir.set_description("Directory containing the auxilliary database");
                optauxdir.set_arg_description("DIR");
                optdisableaux.set_short_name('A');
                optdisableaux.set_long_name("disableaux");
                optdisableaux.set_description("Disable the auxilliary database");
                optdisableaux.set_arg_description("BOOL");
                optgroup.add_entry_filename(optmaindir, dir_main);
                optgroup.add_entry_filename(optauxdir, dir_aux);
                optgroup.add_entry(optdisableaux, dis_aux);
                Glib::OptionContext optctx("[--maindir=<dir>] [--auxdir=<dir>] [--disableaux]");
                optctx.set_help_enabled(true);
                optctx.set_ignore_unknown_options(false);
                optctx.set_main_group(optgroup);
                Gtk::Main m(argc, argv, optctx);
                if (dis_aux)
                        auxdbmode = Engine::auxdb_none;
                else if (!dir_aux.empty())
                        auxdbmode = Engine::auxdb_override;
#ifdef HAVE_OSSO
                osso_context_t* osso_context = osso_initialize("vfrairspaceeditor", VERSION, TRUE /* deprecated parameter */, 0 /* Use default Glib main loop context */);
                if (!osso_context) {
                        std::cerr << "osso_initialize() failed." << std::endl;
                        return OSSO_ERROR;
                }
#endif
#ifdef HAVE_HILDON
                Hildon::init();
#endif
                Glib::set_application_name("VFR Airspace Editor");
                Glib::thread_init();
                AirspaceEditor editor(dir_main, auxdbmode, dir_aux);

                Gtk::Main::run(*editor.get_mainwindow());
#ifdef HAVE_OSSO
                osso_deinitialize(osso_context);
#endif
        } catch (const Glib::Exception& e) {
                std::cerr << "glib exception: " << e.what() << std::endl;
                return 1;
        } catch (const std::exception& e) {
                std::cerr << "exception: " << e.what() << std::endl;
                return 1;
        }
        return 0;
}
