/***************************************************************************
 *   Copyright (C) 2007, 2009 by Thomas Sailer                             *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"
#include "engine.h"
#include "baro.h"

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/types.h>
#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif

Engine::Engine(const std::string & dir_main, auxdb_mode_t auxdbmode, const std::string & dir_aux)
        : m_prefs(), m_dbthread(dir_main, get_aux_dir(auxdbmode, dir_aux)), m_fplandb()
{
        // m_fplandb opens in its constructor
        try {
                m_tracksdb.open(dir_main);
        } catch (const std::exception& e) {
                std::cerr << "Error opening tracks database: " << e.what() << std::endl;
        }
}

std::string Engine::get_aux_dir(auxdb_mode_t auxdbmode, const std::string& dir_aux)
{
        switch (auxdbmode) {
                case auxdb_prefs:
                        return (Glib::ustring)m_prefs.globaldbdir;

                case auxdb_override:
                        return dir_aux;

                default:
                        return std::string();
        }
}

std::string Engine::get_default_aux_dir(void)
{
#ifdef __MINGW_H
        {
                gchar *instdir(g_win32_get_package_installation_directory_of_module(0));
                if (instdir)
                        return Glib::build_filename(instdir, "db");
        }
#endif /* __MINGW_H */
        return PACKAGE_DATA_DIR;
}

const Preferences::PreferencesValueInfo<FPlanRoute::id_t, long long> Preferences::curplan_info("curplan", -1);
const Preferences::PreferencesValueInfo<double> Preferences::vcruise_info("vcruise", 110.0);
const Preferences::PreferencesValueInfo<double> Preferences::windspeed_info("windspeed", 0.0);
const Preferences::PreferencesValueInfo<double> Preferences::winddir_info("winddir", 0.0);
const Preferences::PreferencesValueInfo<double> Preferences::qnh_info("qnh", 1013.0);
const Preferences::PreferencesValueInfo<double> Preferences::temp_info("temp", 288.15 - IcaoAtmosphere::degc_to_kelvin);
const Preferences::PreferencesValueInfo<bool, long long> Preferences::gps_info("gps", true);
const Preferences::PreferencesValueInfo<bool, long long> Preferences::autowpt_info("autowpt", true);
const Preferences::PreferencesValueInfo<double> Preferences::vblock_info("vblock", 40.0);
const Preferences::PreferencesValueInfo<int, long long> Preferences::maxalt_info("maxalt", -1);
const Preferences::PreferencesValueInfo<int, long long> Preferences::mapflags_info("mapflags", 0xfff); // MapRenderer::drawflags_all
const Preferences::PreferencesValueInfo<double> Preferences::mapscale_info("mapscale", 0.1);
const Preferences::PreferencesValueInfo<double> Preferences::mapscaleairportdiagram_info("mapscaleairportdiagram", 0.01);
const Preferences::PreferencesValueInfo<Glib::ustring> Preferences::globaldbdir_info("globaldbdir", Engine::get_default_aux_dir());
const Preferences::PreferencesValueInfo<Glib::ustring> Preferences::gpshost_info("gpshost", "localhost");
#ifdef HAVE_GYPSY
const Preferences::PreferencesValueInfo<Glib::ustring> Preferences::gpsport_info("gpsport", "/dev/ttyUSB0");
#else
const Preferences::PreferencesValueInfo<Glib::ustring> Preferences::gpsport_info("gpsport", "2947");
#endif

const std::string Preferences::curplan_name = "curplan";
const std::string Preferences::vcruise_name = "vcruise";
const std::string Preferences::windspeed_name = "windspeed";
const std::string Preferences::winddir_name = "winddir";
const std::string Preferences::qnh_name = "qnh";
const std::string Preferences::temp_name = "temp";
const std::string Preferences::gps_name = "gps";
const std::string Preferences::autowpt_name = "autowpt";
const std::string Preferences::vblock_name = "vblock";
const std::string Preferences::maxalt_name = "maxalt";
const std::string Preferences::mapflags_name = "mapflags";
const std::string Preferences::mapscale_name = "mapscale";
const std::string Preferences::mapscaleairportdiagram_name = "mapscaleairportdiagram";
const std::string Preferences::globaldbdir_name = "globaldbdir";
const std::string Preferences::gpshost_name = "gpshost";
const std::string Preferences::gpsport_name = "gpsport";

const FPlanRoute::id_t Preferences::curplan_dflt = -1;
const double Preferences::vcruise_dflt = 110.0;
const double Preferences::windspeed_dflt = 0.0;
const double Preferences::winddir_dflt = 0.0;
const double Preferences::qnh_dflt = 1013.0;
const double Preferences::temp_dflt = 288.15 - IcaoAtmosphere::degc_to_kelvin;
const bool Preferences::gps_dflt = true;
const bool Preferences::autowpt_dflt = true;
const double Preferences::vblock_dflt = 40.0;
const int Preferences::maxalt_dflt = -1;
const int Preferences::mapflags_dflt = 0xfff; // MapRenderer::drawflags_all
const double Preferences::mapscale_dflt = 0.1;
const double Preferences::mapscaleairportdiagram_dflt = 0.01;
const Glib::ustring Preferences::globaldbdir_dflt = PACKAGE_DATA_DIR;
const Glib::ustring Preferences::gpshost_dflt = "localhost";
#ifdef HAVE_GYPSY
const Glib::ustring Preferences::gpsport_dflt = "/dev/ttyUSB0";
#else
const Glib::ustring Preferences::gpsport_dflt = "2947";
#endif

template <typename T>
void Preferences::PreferencesValue<T>::operator=(datatype_t v)
{
        if (v == m_value)
                return;
        m_value = v;
        m_dirty = true;
        m_change(m_value);
}

Preferences::Preferences(void)
        : curplan(curplan_info),
          vcruise(vcruise_info),
          windspeed(windspeed_info),
          winddir(winddir_info),
          qnh(qnh_info),
          temp(temp_info),
          gps(gps_info),
          autowpt(autowpt_info),
          vblock(vblock_info),
          maxalt(maxalt_info),
          mapflags(mapflags_info),
          mapscale(mapscale_info),
          mapscaleairportdiagram(mapscaleairportdiagram_info),
          globaldbdir(globaldbdir_info),
          gpshost(gpshost_info),
          gpsport(gpsport_info)
{
        Glib::ustring dbname(Glib::build_filename(FPlan::get_userdbpath(), "prefs.db"));
        db.open(dbname);
        {
                sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS prefs (ID TEXT PRIMARY KEY NOT NULL, VALUE);");
                cmd.executenonquery();
        }
        load(curplan);
        load(vcruise);
        load(windspeed);
        load(winddir);
        load(qnh);
        load(temp);
        load(gps);
        load(autowpt);
        load(vblock);
        load(maxalt);
        load(mapflags);
        load(mapscale);
        load(mapscaleairportdiagram);
        load(globaldbdir);
        load(gpshost);
        load(gpsport);
}

Preferences::~Preferences( void )
{
        try {
                commit();
        } catch (...) {
        }
}

void Preferences::commit( void )
{
        sqlite3x::sqlite3_transaction tran(db, true);
        save(curplan);
        save(vcruise);
        save(windspeed);
        save(winddir);
        save(qnh);
        save(temp);
        save(gps);
        save(autowpt);
        save(vblock);
        save(maxalt);
        save(mapflags);
        save(mapscale);
        save(mapscaleairportdiagram);
        save(globaldbdir);
        save(gpshost);
        save(gpsport);
        tran.commit();
        curplan.clear_dirty();
}

const std::string Preferences::insert_command = "INSERT OR REPLACE INTO prefs (ID,VALUE) VALUES (?,?);";
const std::string Preferences::select_command = "SELECT VALUE FROM prefs WHERE ID = ?;";

void Preferences::save( const std::string & name, long long v )
{
        sqlite3x::sqlite3_command cmd(db, insert_command);
        cmd.bind(1, name);
        cmd.bind(2, v);
        cmd.executenonquery();
        std::cerr << "Preferences::save: name " << name << " value " << v << std::endl;
}

void Preferences::save( const std::string & name, const std::string & v )
{
        sqlite3x::sqlite3_command cmd(db, insert_command);
        cmd.bind(1, name);
        cmd.bind(2, v);
        cmd.executenonquery();
        std::cerr << "Preferences::save: name " << name << " value " << v << std::endl;
}

void Preferences::save( const std::string & name, double v )
{
        sqlite3x::sqlite3_command cmd(db, insert_command);
        cmd.bind(1, name);
        cmd.bind(2, v);
        cmd.executenonquery();
        std::cerr << "Preferences::save: name " << name << " value " << v << std::endl;
}

void Preferences::load( const std::string & name, long long & v )
{
        sqlite3x::sqlite3_command cmd(db, select_command);
        cmd.bind(1, name);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (!cursor.step()) {
                std::cerr << "Preferences::load: name " << name << " value not found" << std::endl;
                return;
        }
        v = cursor.getint64(0);
        std::cerr << "Preferences::load: name " << name << " value " << v << std::endl;
}

void Preferences::load( const std::string & name, std::string & v )
{
        sqlite3x::sqlite3_command cmd(db, select_command);
        cmd.bind(1, name);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (!cursor.step()) {
                std::cerr << "Preferences::load: name " << name << " value not found" << std::endl;
                return;
        }
        v = cursor.getstring(0);
        std::cerr << "Preferences::load: name " << name << " value " << v << std::endl;
}

void Preferences::load( const std::string & name, double & v )
{
        sqlite3x::sqlite3_command cmd(db, select_command);
        cmd.bind(1, name);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (!cursor.step()) {
                std::cerr << "Preferences::load: name " << name << " value not found" << std::endl;
                return;
        }
        v = cursor.getdouble(0);
        std::cerr << "Preferences::load: name " << name << " value " << v << std::endl;
}

#ifdef HAVE_PILOTLINK
PalmDbBase::PalmStringCompare::comp_t Engine::to_palm_compare(DbBaseCommon::comp_t comp)
{
        switch (comp) {
                default:
                case DbBaseCommon::comp_startswith:
                        return PalmDbBase::PalmStringCompare::comp_startswith;

                case DbBaseCommon::comp_contains:
                        return PalmDbBase::PalmStringCompare::comp_contains;

                case DbBaseCommon::comp_exact:
                        return PalmDbBase::PalmStringCompare::comp_exact;
        }
}
#endif

Engine::DbThread::DbThread(const std::string & dir_main, const std::string & dir_aux)
        : m_terminate(false), m_thread(0)
{
        std::cerr << "Database directories: main " << dir_main << " aux " << dir_aux << std::endl;
        try {
                m_airportdb.open(dir_main);
                if (!dir_aux.empty() && m_airportdb.is_dbfile_exists(dir_aux))
                        m_airportdb.attach(dir_aux);
                if (m_airportdb.is_aux_attached())
                        std::cerr << "Auxillary airports database attached" << std::endl;
        } catch (const std::exception& e) {
                std::cerr << "Error opening airports database: " << e.what() << std::endl;
        }
        try {
                m_navaiddb.open(dir_main);
                if (!dir_aux.empty() && m_navaiddb.is_dbfile_exists(dir_aux))
                        m_navaiddb.attach(dir_aux);
                if (m_navaiddb.is_aux_attached())
                        std::cerr << "Auxillary navaids database attached" << std::endl;
        } catch (const std::exception& e) {
                std::cerr << "Error opening navaids database: " << e.what() << std::endl;
        }
        try {
                m_waypointdb.open(dir_main);
                if (!dir_aux.empty() && m_waypointdb.is_dbfile_exists(dir_aux))
                        m_waypointdb.attach(dir_aux);
                if (m_waypointdb.is_aux_attached())
                        std::cerr << "Auxillary waypoints database attached" << std::endl;
        } catch (const std::exception& e) {
                std::cerr << "Error opening waypoints database: " << e.what() << std::endl;
        }
        try {
                m_airspacedb.open(dir_main);
                if (!dir_aux.empty() && m_airspacedb.is_dbfile_exists(dir_aux))
                        m_airspacedb.attach(dir_aux);
                if (m_airspacedb.is_aux_attached())
                        std::cerr << "Auxillary airspaces database attached" << std::endl;
        } catch (const std::exception& e) {
                std::cerr << "Error opening airspaces database: " << e.what() << std::endl;
        }
        try {
                m_airwaydb.open(dir_main);
                if (!dir_aux.empty() && m_airwaydb.is_dbfile_exists(dir_aux))
                        m_airwaydb.attach(dir_aux);
                if (m_airwaydb.is_aux_attached())
                        std::cerr << "Auxillary airways database attached" << std::endl;
        } catch (const std::exception& e) {
                std::cerr << "Error opening airways database: " << e.what() << std::endl;
        }
        try {
                m_mapelementdb.open(dir_aux.empty() ? get_default_aux_dir() : dir_aux);
        } catch (const std::exception& e) {
                std::cerr << "Error opening mapelements database: " << e.what() << std::endl;
        }
        try {
                m_topodb.open(dir_aux.empty() ? get_default_aux_dir() : dir_aux);
        } catch (const std::exception& e) {
                std::cerr << "Error opening topo database: " << e.what() << std::endl;
        }
#ifdef HAVE_PILOTLINK
        try {
                m_palmairportdb.open(Glib::build_filename(get_default_aux_dir(), "AirportsDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm airports database: " << e.what() << std::endl;
        }
        try {
                m_palmnavaiddb.open(Glib::build_filename(get_default_aux_dir(), "NavaidsDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm navaids database: " << e.what() << std::endl;
        }
        try {
                m_palmwaypointdb.open(Glib::build_filename(get_default_aux_dir(), "WaypointsDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm waypoints database: " << e.what() << std::endl;
        }
        try {
                m_palmairspacedb.open(Glib::build_filename(get_default_aux_dir(), "AirspaceDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm airspace database: " << e.what() << std::endl;
        }
        try {
                m_palmmapelementdb.open(Glib::build_filename(get_default_aux_dir(), "MapelementDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm map elements database: " << e.what() << std::endl;
        }
        try {
                m_palmgtopodb.open(Glib::build_filename(get_default_aux_dir(), "GTopoDB.pdb"));
        } catch (const std::exception& e) {
                std::cerr << "Error opening palm gtopo database: " << e.what() << std::endl;
        }
#endif
        m_thread = Glib::Thread::create(sigc::mem_fun(*this, &Engine::DbThread::thread), 0, true, true, Glib::THREAD_PRIORITY_NORMAL);
}

Engine::DbThread::~DbThread()
{
        if (m_thread) {
                m_terminate = true;
                m_cond.broadcast();
                m_thread->join();
                m_thread = 0;
        }
}

void Engine::DbThread::thread( void )
{
        m_mutex.lock();
        while (!m_terminate) {
                if (m_tasklist.empty()) {
                        m_cond.wait(m_mutex);
                        if (m_tasklist.empty())
                                continue;
                }
                Glib::RefPtr<ResultBase> task(m_tasklist.front());
                m_tasklist.pop_front();
                m_mutex.unlock();
                if (!task->is_error())
                        task->execute();
                m_mutex.lock();
        }
        while (!m_tasklist.empty()) {
                Glib::RefPtr<ResultBase> task(m_tasklist.front());
                m_tasklist.pop_front();
                m_mutex.unlock();
                task->seterror();
                m_mutex.lock();
        }
        m_mutex.unlock();
}

template<class T> void Engine::DbThread::queue(Glib::RefPtr<T>& p)
{
        if (!p)
                return;
        if (m_terminate) {
                p->seterror();
                return;
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                m_tasklist.push_back(p);
                m_cond.broadcast();
        }
}

Engine::DbThread::ResultBase::ResultBase(const sigc::slot<void> & dbaction, const sigc::slot<void>& dbcancel)
        : m_dbcancel(dbcancel), m_dbaction(dbaction), m_refcount(1), m_done(false), m_error(false), m_executing(false)
{
#ifdef HAVE_SYS_SYSCALL_H
        //std::cerr << "Engine::DbThread::ResultBase::ResultBase " << (unsigned long)this << " TID " << syscall(SYS_gettid) << std::endl;
#endif
}

Engine::DbThread::ResultBase::~ResultBase()
{
#ifdef HAVE_SYS_SYSCALL_H
        //std::cerr << "Engine::DbThread::ResultBase::~ResultBase " << (unsigned long)this << " TID " << syscall(SYS_gettid) << std::endl;
#endif
}

void Engine::DbThread::ResultBase::reference(void )
{
        Glib::Mutex::Lock lock(m_mutex);
        ++m_refcount;
}

void Engine::DbThread::ResultBase::unreference(void )
{
        int rcnt;
        {
                Glib::Mutex::Lock lock(m_mutex);
                rcnt = --m_refcount;
        }
        if (!rcnt)
                delete this;
}

sigc::connection Engine::DbThread::ResultBase::connect(const sigc::signal< void >::slot_type & slot)
{
        sigc::connection conn;
        bool done;
        {
                Glib::Mutex::Lock lock(m_mutex);
                done = m_done;
                conn = m_donecb.connect(slot);
        }
        if (done)
                slot();
        return conn;
}

void Engine::DbThread::ResultBase::disconnect_all(void )
{
        Glib::Mutex::Lock lock(m_mutex);
        m_donecb.clear();
}

void Engine::DbThread::ResultBase::cancel(void )
{
        Glib::Mutex::Lock lock(m_mutex);
        m_done = true;
        m_error = true;
        if (m_executing)
                m_dbcancel();
}

template<class T> void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<T> & q)
{
        if (q) {
                {
                        Glib::Mutex::Lock lock(q->m_mutex);
                        q->m_donecb.clear();
                        if (!q->is_done()) {
                                q->m_done = true;
                                q->m_error = true;
                                if (q->m_executing)
                                        q->m_dbcancel();
                        }
                }
                q = Glib::RefPtr<T>();
        }
}

template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<AirportResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<AirspaceResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<NavaidResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<WaypointResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<AirwayResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<MapelementResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<ElevationResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<ElevationMinMaxResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<ElevationMapResult> & q);
template void Engine::DbThread::ResultBase::cancel(Glib::RefPtr<ElevationMapCairoResult> & q);

void Engine::DbThread::ResultBase::execute(void)
{
        //std::cerr << "Engine::DbThread::ResultBase::execute " << (unsigned long)this << std::endl;
        if (!m_error) {
                m_executing = true;
                try {
                        m_dbaction();
                } catch (const std::exception& e) {
                        std::cerr << "DbThread::execute: exception " << e.what() << std::endl;
                        m_error = true;
                }
        }
        m_executing = false;
        //std::cerr << "Engine::DbThread::ResultBase::execute done " << (unsigned long)this << std::endl;
        Glib::Mutex::Lock lock(m_mutex);
        m_done = true;
        m_donecb();
}

void Engine::DbThread::ResultBase::seterror( void )
{
        //std::cerr << "Engine::DbThread::ResultBase::seterror " << (unsigned long)this << std::endl;
        Glib::Mutex::Lock lock(m_mutex);
        m_done = true;
        m_error = true;
        if (m_executing)
                m_dbcancel();
}

template<class Db> Engine::DbThread::Result<Db>::Result(const sigc::slot<elementvector_t> & dbaction, const sigc::slot<void>& dbcancel)
        : ResultBase(sigc::compose(sigc::mem_fun(*this, &Engine::DbThread::Result<Db>::set_result), dbaction), dbcancel)
{
}

template<class Db> Engine::DbThread::Result<Db>::~Result()
{
}

template class Engine::DbThread::Result<NavaidsDb>;
template class Engine::DbThread::Result<WaypointsDb>;
template class Engine::DbThread::Result<AirwaysDb>;
template class Engine::DbThread::Result<AirportsDb>;
template class Engine::DbThread::Result<AirspacesDb>;
template class Engine::DbThread::Result<MapelementsDb>;


Engine::DbThread::ElevationResult::ElevationResult(const Point & pt, TopoDb30 & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationResult::execute), sigc::ref(topodb), pt), sigc::mem_fun(topodb, &TopoDb30::interrupt)),
          m_result(0)
{
}

#ifdef HAVE_PILOTLINK
Engine::DbThread::ElevationResult::ElevationResult(const Point & pt, PalmGtopo & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationResult::execute_palm), sigc::ref(topodb), pt)),
          m_result(0)
{
}
#endif

Engine::DbThread::ElevationResult::~ElevationResult()
{
}

void Engine::DbThread::ElevationResult::execute(TopoDb30& topodb, const Point& pt)
{
        m_result = topodb.get_elev(pt);
}

#ifdef HAVE_PILOTLINK
void Engine::DbThread::ElevationResult::execute_palm(PalmGtopo& topodb, const Point& pt)
{
        m_result = topodb[pt];
}
#endif

Engine::DbThread::ElevationMinMaxResult::ElevationMinMaxResult(const Rect & r, TopoDb30 & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMinMaxResult::execute_rect), sigc::ref(topodb), r), sigc::mem_fun(topodb, &TopoDb30::interrupt)),
          m_result(TopoDb30::nodata, TopoDb30::nodata)
{
}

Engine::DbThread::ElevationMinMaxResult::ElevationMinMaxResult(const Polygon & p, TopoDb30 & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMinMaxResult::execute_poly), sigc::ref(topodb), p), sigc::mem_fun(topodb, &TopoDb30::interrupt)),
          m_result(TopoDb30::nodata, TopoDb30::nodata)
{
}

#ifdef HAVE_PILOTLINK
Engine::DbThread::ElevationMinMaxResult::ElevationMinMaxResult(const Rect & r, PalmGtopo & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMinMaxResult::execute_rect_palm), sigc::ref(topodb), r)),
          m_result(TopoDb30::nodata, TopoDb30::nodata)
{
}

Engine::DbThread::ElevationMinMaxResult::ElevationMinMaxResult(const Polygon & p, PalmGtopo & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMinMaxResult::execute_poly_palm), sigc::ref(topodb), p)),
          m_result(TopoDb30::nodata, TopoDb30::nodata)
{
}
#endif

Engine::DbThread::ElevationMinMaxResult::~ElevationMinMaxResult()
{
}

#ifdef HAVE_PILOTLINK
void Engine::DbThread::ElevationMinMaxResult::execute_poly_palm(PalmGtopo & topodb, const Polygon & p)
{
        m_result = topodb.get_minmax_elev(p);
}

void Engine::DbThread::ElevationMinMaxResult::execute_rect_palm(PalmGtopo & topodb, const Rect & r)
{
        m_result = topodb.get_minmax_elev(r);
}
#endif

void Engine::DbThread::ElevationMinMaxResult::execute_poly(TopoDb30 & topodb, const Polygon & p)
{
        m_result = topodb.get_minmax_elev(p);
}

void Engine::DbThread::ElevationMinMaxResult::execute_rect(TopoDb30 & topodb, const Rect & r)
{
        m_result = topodb.get_minmax_elev(r);
}

Engine::DbThread::ElevationMapResult::ElevationMapResult(const Rect & bbox, TopoDb30 & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMapResult::execute), sigc::ref(topodb)), sigc::mem_fun(topodb, &TopoDb30::interrupt)),
          m_bbox(bbox), m_width(0), m_height(0), m_elev(0)
{
}

#ifdef HAVE_PILOTLINK
Engine::DbThread::ElevationMapResult::ElevationMapResult(const Rect & bbox, PalmGtopo & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMapResult::execute_palm), sigc::ref(topodb))),
          m_bbox(bbox), m_width(0), m_height(0), m_elev(0)
{
}
#endif

Engine::DbThread::ElevationMapResult::~ElevationMapResult()
{
        delete[] m_elev;
}

TopoDb30::elev_t Engine::DbThread::ElevationMapResult::operator()(unsigned int x, unsigned int y)
{
        if (!m_elev || x >= m_width || y >= m_height)
                return TopoDb30::nodata;
        return m_elev[y * m_width + x];
}

void Engine::DbThread::ElevationMapResult::execute(TopoDb30& topodb)
{
        TopoDb30::TopoCoordinate gmin(m_bbox.get_southwest()), gmax(m_bbox.get_northeast());
        gmin.advance_west();
        gmin.advance_south();
        gmax.advance_east();
        gmax.advance_north();
        Point pixeldim(TopoDb30::TopoCoordinate::get_pointsize());
        std::cerr << "TopoDb: from " << gmin << " to " << gmax << std::endl;
        unsigned int pbufx(gmin.lon_dist(gmax)+1);
        unsigned int pbufy(gmin.lat_dist(gmax)+1);
        delete[] m_elev;
        m_elev = 0;
        m_width = 0;
        m_height = 0;
        if ((pbufx * pbufy) > (1024 * 1024)) {
                seterror();
                return;
        }
        m_elev = new TopoDb30::elev_t[pbufx * pbufy];
        if (!m_elev) {
                seterror();
                return;
        }
        m_width = pbufx;
        m_height = pbufy;
        for (unsigned int i = 0; i < pbufx * pbufy; i++)
                m_elev[i] = 0x5555;
        TopoDb30::TopoTileCoordinate tmin(gmin), tmax(gmax);
        unsigned int pboffsx(0), pboffsy(0);
        for (TopoDb30::TopoTileCoordinate tc(tmin);;) {
                TopoDb30::pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                TopoDb30::pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : TopoDb30::tile_size - 1);
                TopoDb30::pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                TopoDb30::pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : TopoDb30::tile_size - 1);
                xmax -= xmin;
                ymax -= ymin;
                for (TopoDb30::pixel_index_t y(0); y <= ymax; y++) {
                        tc.set_lat_offs(ymin + y);
                        for (TopoDb30::pixel_index_t x(0); x <= xmax; x++) {
                                tc.set_lon_offs(xmin + x);
                                TopoDb30::elev_t elev(topodb.get_elev(tc));
                                //std::cerr << "TopoDb: point " << (TopoDb30::TopoCoordinate)tc << ' ' << (Point)(TopoDb30::TopoCoordinate)tc << " elev " << elev << std::endl;
                                unsigned int xx(pboffsx + x), yy(pbufy - 1 - pboffsy - y);
                                if (xx < m_width && yy < m_height)
                                        m_elev[yy * m_width + xx] = elev;
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        pboffsx += xmax + 1;
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                pboffsx = 0;
                tc.advance_tile_north();
                pboffsy += ymax + 1;
        }
        m_bbox = Rect((TopoDb30::TopoCoordinate)tmin, pixeldim + (TopoDb30::TopoCoordinate)tmax);
}

#ifdef HAVE_PILOTLINK
void Engine::DbThread::ElevationMapResult::execute_palm(PalmGtopo& topodb)
{
        PalmGtopo::GtopoCoord gmin(m_bbox.get_southwest()), gmax(m_bbox.get_northeast());
        gmin -= PalmGtopo::GtopoCoord(1, 1);
        gmax += PalmGtopo::GtopoCoord(1, 1);
        std::cerr << "Gtopo: from " << gmin << " to " << gmax << std::endl;
        unsigned int pbufx((gmax-gmin).get_lon()+1);
        unsigned int pbufy((gmax-gmin).get_lat()+1);
        delete[] m_elev;
        m_elev = 0;
        m_width = 0;
        m_height = 0;
        if ((pbufx * pbufy) > (1024 * 1024)) {
                seterror();
                return;
        }
        m_elev = new TopoDb30::elev_t[pbufx * pbufy];
        if (!m_elev) {
                seterror();
                return;
        }
        m_width = pbufx;
        m_height = pbufy;
        for (PalmGtopo::GtopoCoord gp(gmin);;) {
                //std::cerr << "Gtopo: point " << gp << std::endl;
                PalmGtopo::elevation_t elev(topodb[gp]);
                PalmGtopo::GtopoCoord gidx(gp - gmin);
                unsigned int xx(gidx.get_lon()), yy(pbufy - 1 - gidx.get_lat());
                if (xx < m_width && yy < m_height)
                        m_elev[yy * m_width + xx] = elev;
                if (gp == gmax)
                        break;
                if (gp.get_lon() != gmax.get_lon()) {
                        gp.set_lon(gp.get_lon()+1);
                        continue;
                }
                gp.set_lon(gmin.get_lon());
                gp.set_lat(gp.get_lat()+1);
        }
        m_bbox = Rect(gmin, gmax + PalmGtopo::GtopoCoord(1, 1));
}
#endif

Engine::DbThread::ElevationMapCairoResult::ElevationMapCairoResult(const Rect & bbox, TopoDb30 & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMapCairoResult::execute), sigc::ref(topodb)), sigc::mem_fun(topodb, &TopoDb30::interrupt)),
          m_bbox(bbox)
{
}

#ifdef HAVE_PILOTLINK
Engine::DbThread::ElevationMapCairoResult::ElevationMapCairoResult(const Rect & bbox, PalmGtopo & topodb)
        : ResultBase(sigc::bind(sigc::mem_fun(*this, &Engine::DbThread::ElevationMapCairoResult::execute_palm), sigc::ref(topodb))),
          m_bbox(bbox)
{
}
#endif

Engine::DbThread::ElevationMapCairoResult::~ElevationMapCairoResult()
{
}

void Engine::DbThread::ElevationMapCairoResult::execute(TopoDb30& topodb)
{
        TopoDb30::TopoCoordinate gmin(m_bbox.get_southwest()), gmax(m_bbox.get_northeast());
        gmin.advance_west();
        gmin.advance_south();
        gmax.advance_east();
        gmax.advance_north();
        Point pixeldim(TopoDb30::TopoCoordinate::get_pointsize());
        std::cerr << "TopoDb: from " << gmin << " to " << gmax << std::endl;
        unsigned int pbufx(gmin.lon_dist(gmax)+1);
        unsigned int pbufy(gmin.lat_dist(gmax)+1);
        m_surface = Cairo::ImageSurface::create(Cairo::FORMAT_RGB24, pbufx, pbufy);
        unsigned char *pbuf_data(m_surface->get_data());
        int pbuf_stride(m_surface->get_stride());
        TopoDb30::TopoTileCoordinate tmin(gmin), tmax(gmax);
        unsigned int pboffsx(0), pboffsy(0);
        for (TopoDb30::TopoTileCoordinate tc(tmin);;) {
                TopoDb30::pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                TopoDb30::pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : TopoDb30::tile_size - 1);
                TopoDb30::pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                TopoDb30::pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : TopoDb30::tile_size - 1);
                xmax -= xmin;
                ymax -= ymin;
                for (TopoDb30::pixel_index_t y(0); y <= ymax; y++) {
                        tc.set_lat_offs(ymin + y);
                        for (TopoDb30::pixel_index_t x(0); x <= xmax; x++) {
                                tc.set_lon_offs(xmin + x);
                                TopoDb30::elev_t elev(topodb.get_elev(tc));
                                //std::cerr << "TopoDb: point " << (TopoDb30::TopoCoordinate)tc << ' ' << (Point)(TopoDb30::TopoCoordinate)tc << " elev " << elev << std::endl;
                                if (elev == TopoDb30::nodata) {
                                        ((TopoDb30::TopoCoordinate)tc).print(std::cerr << "No topo data for " << (Point)(TopoDb30::TopoCoordinate)tc << " TC ") << " ti " << tc.get_tile_index() << " pi " << tc.get_pixel_index() << std::endl;
                                } else {
                                        unsigned char *d = pbuf_data + 4 * (pboffsx + x) + pbuf_stride * (pbufy - 1 - pboffsy - y);
                                        uint32_t *d32((uint32_t *)d);
                                        *d32 = TopoDb30::color(elev);
                                        //std::cerr << "Pixel X " << (pboffsx + x) << " Y " << (pboffsy + y) << " elev " << elev << std::endl;
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        pboffsx += xmax + 1;
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                pboffsx = 0;
                tc.advance_tile_north();
                pboffsy += ymax + 1;
        }
        m_bbox = Rect((TopoDb30::TopoCoordinate)tmin, pixeldim + (TopoDb30::TopoCoordinate)tmax);
}

#ifdef HAVE_PILOTLINK
void Engine::DbThread::ElevationMapCairoResult::execute_palm(PalmGtopo& topodb)
{
        PalmGtopo::GtopoCoord gmin(m_bbox.get_southwest()), gmax(m_bbox.get_northeast());
        gmin -= PalmGtopo::GtopoCoord(1, 1);
        gmax += PalmGtopo::GtopoCoord(1, 1);
        std::cerr << "Gtopo: from " << gmin << " to " << gmax << std::endl;
        unsigned int pbufx((gmax-gmin).get_lon()+1);
        unsigned int pbufy((gmax-gmin).get_lat()+1);
        m_surface = Cairo::ImageSurface::create(Cairo::FORMAT_RGB24, pbufx, pbufy);
        unsigned char *pbuf_data(m_surface->get_data());
        int pbuf_stride(m_surface->get_stride());
        for (PalmGtopo::GtopoCoord gp(gmin);;) {
                //std::cerr << "Gtopo: point " << gp << std::endl;
                PalmGtopo::elevation_t elev(topodb[gp]);
                PalmGtopo::GtopoCoord gidx(gp - gmin);
                unsigned char *d = pbuf_data + 4 * gidx.get_lon() + pbuf_stride * (pbufy - 1 - gidx.get_lat());
                uint32_t *d32((uint32_t *)d);
                *d32 = TopoDb30::color(elev);
                if (gp == gmax)
                        break;
                if (gp.get_lon() != gmax.get_lon()) {
                        gp.set_lon(gp.get_lon()+1);
                        continue;
                }
                gp.set_lon(gmin.get_lon());
                gp.set_lat(gp.get_lat()+1);
        }
        m_bbox = Rect(gmin, gmax + PalmGtopo::GtopoCoord(1, 1));
}
#endif

Glib::RefPtr<Engine::DbThread::AirportResult> Engine::DbThread::airport_find_by_icao(const Glib::ustring & nm, unsigned int limit, unsigned int skip, AirportsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirportResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairportdb.is_open())
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_palmairportdb, &PalmAirports::find_by_icao), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_airportdb, &AirportsDb::find_by_icao), nm, 0, comp, limit, loadsubtables),
                                                                  sigc::mem_fun(m_airportdb, &AirportsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirportResult> Engine::DbThread::airport_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, AirportsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirportResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairportdb.is_open())
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_palmairportdb, &PalmAirports::find_by_name), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_airportdb, &AirportsDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                                                  sigc::mem_fun(m_airportdb, &AirportsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirportResult> Engine::DbThread::airport_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<AirportResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairportdb.is_open())
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_palmairportdb, &PalmAirports::find_bbox), rect, limit)));
        else
#endif
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_airportdb, &AirportsDb::find_by_rect), rect, limit, loadsubtables),
                                                                  sigc::mem_fun(m_airportdb, &AirportsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirportResult> Engine::DbThread::airport_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<AirportResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairportdb.is_open())
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_palmairportdb, &PalmAirports::find_nearest), pt, limit, rect)));
        else
#endif
                p = Glib::RefPtr<AirportResult>(new AirportResult(sigc::bind(sigc::mem_fun(m_airportdb, &AirportsDb::find_nearest), pt, rect, limit, loadsubtables),
                                                                  sigc::mem_fun(m_airportdb, &AirportsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirportSaveResult> Engine::DbThread::airport_save(const AirportsDb::Airport & e)
{
        Glib::RefPtr<AirportSaveResult> p(new AirportSaveResult(sigc::bind(sigc::mem_fun(m_airportdb, &AirportsDb::save), e),
                                                                sigc::mem_fun(m_airportdb, &AirportsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirspaceResult> Engine::DbThread::airspace_find_by_icao(const Glib::ustring & nm, unsigned int limit, unsigned int skip, AirspacesDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirspaceResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairspacedb.is_open())
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_palmairspacedb, &PalmAirspaces::find_by_icao), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_airspacedb, &AirspacesDb::find_by_icao), nm, 0, comp, limit, loadsubtables),
                                                                    sigc::mem_fun(m_airspacedb, &AirspacesDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirspaceResult> Engine::DbThread::airspace_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, AirspacesDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirspaceResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairspacedb.is_open())
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_palmairspacedb, &PalmAirspaces::find_by_name), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_airspacedb, &AirspacesDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                                                    sigc::mem_fun(m_airspacedb, &AirspacesDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirspaceResult> Engine::DbThread::airspace_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<AirspaceResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairspacedb.is_open())
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_palmairspacedb, &PalmAirspaces::find_bbox), rect, limit)));
        else
#endif
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_airspacedb, &AirspacesDb::find_by_rect), rect, limit, loadsubtables),
                                                                    sigc::mem_fun(m_airspacedb, &AirspacesDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirspaceResult> Engine::DbThread::airspace_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<AirspaceResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmairspacedb.is_open())
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_palmairspacedb, &PalmAirspaces::find_nearest), pt, limit, rect)));
        else
#endif
                p = Glib::RefPtr<AirspaceResult>(new AirspaceResult(sigc::bind(sigc::mem_fun(m_airspacedb, &AirspacesDb::find_nearest), pt, rect, limit, loadsubtables),
                                                                    sigc::mem_fun(m_airspacedb, &AirspacesDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirspaceSaveResult> Engine::DbThread::airspace_save(const AirspacesDb::Airspace & e)
{
        Glib::RefPtr<AirspaceSaveResult> p(new AirspaceSaveResult(sigc::bind(sigc::mem_fun(m_airspacedb, &AirspacesDb::save), e),
                                                                  sigc::mem_fun(m_airspacedb, &AirspacesDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::NavaidResult> Engine::DbThread::navaid_find_by_icao(const Glib::ustring & nm, unsigned int limit, unsigned int skip, NavaidsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<NavaidResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmnavaiddb.is_open())
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_palmnavaiddb, &PalmNavaids::find_by_icao), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_navaiddb, &NavaidsDb::find_by_icao), nm, 0, comp, limit, loadsubtables),
                                                                sigc::mem_fun(m_navaiddb, &NavaidsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::NavaidResult> Engine::DbThread::navaid_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, NavaidsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<NavaidResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmnavaiddb.is_open())
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_palmnavaiddb, &PalmNavaids::find_by_name), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_navaiddb, &NavaidsDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                                                sigc::mem_fun(m_navaiddb, &NavaidsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::NavaidResult> Engine::DbThread::navaid_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<NavaidResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmnavaiddb.is_open())
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_palmnavaiddb, &PalmNavaids::find_bbox), rect, limit)));
        else
#endif
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_navaiddb, &NavaidsDb::find_by_rect), rect, limit, loadsubtables),
                                                                sigc::mem_fun(m_navaiddb, &NavaidsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::NavaidResult> Engine::DbThread::navaid_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<NavaidResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmnavaiddb.is_open())
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_palmnavaiddb, &PalmNavaids::find_nearest), pt, limit, rect)));
        else
#endif
                p = Glib::RefPtr<NavaidResult>(new NavaidResult(sigc::bind(sigc::mem_fun(m_navaiddb, &NavaidsDb::find_nearest), pt, rect, limit, loadsubtables),
                                                                sigc::mem_fun(m_navaiddb, &NavaidsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::NavaidSaveResult> Engine::DbThread::navaid_save(const NavaidsDb::Navaid & e)
{
        Glib::RefPtr<NavaidSaveResult> p(new NavaidSaveResult(sigc::bind(sigc::mem_fun(m_navaiddb, &NavaidsDb::save), e),
                                                              sigc::mem_fun(m_navaiddb, &NavaidsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::WaypointResult> Engine::DbThread::waypoint_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, WaypointsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<WaypointResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmwaypointdb.is_open())
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_palmwaypointdb, &PalmWaypoints::find_by_name), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_waypointdb, &WaypointsDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                                                    sigc::mem_fun(m_waypointdb, &WaypointsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::WaypointResult> Engine::DbThread::waypoint_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<WaypointResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmwaypointdb.is_open())
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_palmwaypointdb, &PalmWaypoints::find_bbox), rect, limit)));
        else
#endif
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_waypointdb, &WaypointsDb::find_by_rect), rect, limit, loadsubtables),
                                                                    sigc::mem_fun(m_waypointdb, &WaypointsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::WaypointResult> Engine::DbThread::waypoint_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<WaypointResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmwaypointdb.is_open())
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_palmwaypointdb, &PalmWaypoints::find_nearest), pt, limit, rect)));
        else
#endif
                p = Glib::RefPtr<WaypointResult>(new WaypointResult(sigc::bind(sigc::mem_fun(m_waypointdb, &WaypointsDb::find_nearest), pt, rect, limit, loadsubtables),
                                                                    sigc::mem_fun(m_waypointdb, &WaypointsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::WaypointSaveResult> Engine::DbThread::waypoint_save(const WaypointsDb::Waypoint & e)
{
        Glib::RefPtr<WaypointSaveResult> p(new WaypointSaveResult(sigc::bind(sigc::mem_fun(m_waypointdb, &WaypointsDb::save), e),
                                                                  sigc::mem_fun(m_waypointdb, &WaypointsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirwayResult> Engine::DbThread::airway_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, AirwaysDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirwayResult> p;
        p = Glib::RefPtr<AirwayResult>(new AirwayResult(sigc::bind(sigc::mem_fun(m_airwaydb, &AirwaysDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                       sigc::mem_fun(m_airwaydb, &AirwaysDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirwayResult> Engine::DbThread::airway_find_by_text(const Glib::ustring& nm, unsigned int limit, unsigned int skip, AirwaysDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<AirwayResult> p;
        p = Glib::RefPtr<AirwayResult>(new AirwayResult(sigc::bind(sigc::mem_fun(m_airwaydb, &AirwaysDb::find_by_text), nm, 0, comp, limit, loadsubtables),
                                       sigc::mem_fun(m_airwaydb, &AirwaysDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirwayResult> Engine::DbThread::airway_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<AirwayResult> p;
        p = Glib::RefPtr<AirwayResult>(new AirwayResult(sigc::bind(sigc::mem_fun(m_airwaydb, &AirwaysDb::find_by_rect), rect, limit, loadsubtables),
                                       sigc::mem_fun(m_airwaydb, &AirwaysDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirwayResult> Engine::DbThread::airway_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<AirwayResult> p;
        p = Glib::RefPtr<AirwayResult>(new AirwayResult(sigc::bind(sigc::mem_fun(m_airwaydb, &AirwaysDb::find_nearest), pt, rect, limit, loadsubtables),
                                       sigc::mem_fun(m_airwaydb, &AirwaysDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::AirwaySaveResult> Engine::DbThread::airway_save(const AirwaysDb::Airway & e)
{
        Glib::RefPtr<AirwaySaveResult> p(new AirwaySaveResult(sigc::bind(sigc::mem_fun(m_airwaydb, &AirwaysDb::save), e),
                                         sigc::mem_fun(m_airwaydb, &AirwaysDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::MapelementResult> Engine::DbThread::mapelement_find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, WaypointsDb::comp_t comp, unsigned int loadsubtables)
{
        Glib::RefPtr<MapelementResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmmapelementdb.is_open())
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_palmmapelementdb, &PalmMapelements::find_by_name), nm, limit, skip, to_palm_compare(comp))));
        else
#endif
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_mapelementdb, &MapelementsDb::find_by_name), nm, 0, comp, limit, loadsubtables),
                                                                        sigc::mem_fun(m_mapelementdb, &MapelementsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::MapelementResult> Engine::DbThread::mapelement_find_bbox(const Rect & rect, unsigned int limit, unsigned int loadsubtables)
{
        Glib::RefPtr<MapelementResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmmapelementdb.is_open())
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_palmmapelementdb, &PalmMapelements::find_bbox), rect, limit)));
        else
#endif
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_mapelementdb, &MapelementsDb::find_by_rect), rect, limit, loadsubtables),
                                                                        sigc::mem_fun(m_mapelementdb, &MapelementsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::MapelementResult> Engine::DbThread::mapelement_find_nearest(const Point & pt, unsigned int limit, const Rect & rect, unsigned int loadsubtables)
{
        Glib::RefPtr<MapelementResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmmapelementdb.is_open())
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_palmmapelementdb, &PalmMapelements::find_nearest), pt, limit, rect)));
        else
#endif
                p = Glib::RefPtr<MapelementResult>(new MapelementResult(sigc::bind(sigc::mem_fun(m_mapelementdb, &MapelementsDb::find_nearest), pt, rect, limit, loadsubtables),
                                                                        sigc::mem_fun(m_mapelementdb, &MapelementsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::MapelementSaveResult> Engine::DbThread::mapelement_save(const MapelementsDb::Mapelement & e)
{
        Glib::RefPtr<MapelementSaveResult> p(new MapelementSaveResult(sigc::bind(sigc::mem_fun(m_mapelementdb, &MapelementsDb::save), e),
                                                                      sigc::mem_fun(m_mapelementdb, &MapelementsDb::interrupt)));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::ElevationResult> Engine::DbThread::elevation_point(const Point & pt)
{
        Glib::RefPtr<ElevationResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmgtopodb.is_open())
                p = Glib::RefPtr<ElevationResult>(new ElevationResult(pt, m_palmgtopodb));
        else
#endif
                p = Glib::RefPtr<ElevationResult>(new ElevationResult(pt, m_topodb));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::ElevationMapResult> Engine::DbThread::elevation_map(const Rect & bbox)
{
        Glib::RefPtr<ElevationMapResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmgtopodb.is_open())
                p = Glib::RefPtr<ElevationMapResult>(new ElevationMapResult(bbox, m_palmgtopodb));
        else
#endif
                p = Glib::RefPtr<ElevationMapResult>(new ElevationMapResult(bbox, m_topodb));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::ElevationMapCairoResult> Engine::DbThread::elevation_map_cairo(const Rect & bbox)
{
        Glib::RefPtr<ElevationMapCairoResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmgtopodb.is_open())
                p = Glib::RefPtr<ElevationMapCairoResult>(new ElevationMapCairoResult(bbox, m_palmgtopodb));
        else
#endif
                p = Glib::RefPtr<ElevationMapCairoResult>(new ElevationMapCairoResult(bbox, m_topodb));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::ElevationMinMaxResult> Engine::DbThread::elevation_minmax(const Rect & r)
{
        Glib::RefPtr<ElevationMinMaxResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmgtopodb.is_open())
                p = Glib::RefPtr<ElevationMinMaxResult>(new ElevationMinMaxResult(r, m_palmgtopodb));
        else
#endif
                p = Glib::RefPtr<ElevationMinMaxResult>(new ElevationMinMaxResult(r, m_topodb));
        queue(p);
        return p;
}

Glib::RefPtr<Engine::DbThread::ElevationMinMaxResult> Engine::DbThread::elevation_minmax(const Polygon & po)
{
        Glib::RefPtr<ElevationMinMaxResult> p;
#ifdef HAVE_PILOTLINK
        if (m_palmgtopodb.is_open())
                p = Glib::RefPtr<ElevationMinMaxResult>(new ElevationMinMaxResult(po, m_palmgtopodb));
        else
#endif
                p = Glib::RefPtr<ElevationMinMaxResult>(new ElevationMinMaxResult(po, m_topodb));
        queue(p);
        return p;
}
